#include "avs.h"

#include "calculo.h"
#include "comparisons.h"
#include "debugInfo.h"

#include <math.h>
#include <stdlib.h>

/**
 * @file avs.c
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

void GrabarVientoAVS(const char *fich, const struct T_Malla *malla)
{
    FILE *f;

    if ( (f=fopen(fich,"w")) == NULL){
        fprintf(stderr, "No se pudo abrir el fichero");
        fprintf(stderr, "Nombre del fichero %s\n", fich);
        exit(EXIT_FAILURE);
    }

    // Graba el numero de nodos, numero de tetraedros, componentes de
    // datos de los nodos, componentes de datos de los tetraedros.
    // Cada tetraedro tiene asociado un escalar (modulo de la velocidad
    // del viento en el centro del tetraedro) y un vector de 3D (velocidad
    // del viento en el centro del tetraedro)
    fprintf(f, "%zu %zu 0 3 0\n", malla->Num_nodos, malla->Num_tetra);
    //	 Graba las coordenadas de los nodos
    for(size_t i=0; i < malla->Num_nodos; i++) {
        fprintf(f, "%zu %8.6f %8.6f %8.6f\n",
                i, malla->nodo[i][0], malla->nodo[i][1], malla->nodo[i][2]);
    }
    // Graba los tetraedros
    // num. ref., material, tipo, vertices
    for(size_t i=0; i < malla->Num_tetra; i++) {
        fprintf(f, "%zu 1 tet %zu %zu %zu %zu\n",
                i, malla->tetra[i][0], malla->tetra[i][1],
                malla->tetra[i][2], malla->tetra[i][3]);
    }

    // Para los tetraedros: num.comp., tamano comp1, tamano comp2, ...
    fprintf(f, "1 3\n");
    fprintf(f, "V, m/s\n");
    // Graba los datos de los tetraedros: vector velocidad y modulo
    for(size_t i=0; i < malla->Num_tetra; i++) {
        fprintf(f, "%zu %14.10f %14.10f %14.10f\n", i,
                malla->wind[i][0], malla->wind[i][1], malla->wind[i][2]);
    }

    fclose(f);
}

void GrabarVientoInter(const char *fich, const struct T_Malla *malla)
{
    FILE *f;

    if ( (f=fopen(fich,"w")) == NULL){
        fprintf(stderr, "No se pudo abrir el fichero");
        fprintf(stderr, "Nombre del fichero %s\n", fich);
        exit(EXIT_FAILURE);
    }

    // Graba el numero de nodos, numero de tetraedros, componentes de
    // datos de los nodos, componentes de datos de los tetraedros.
    // Cada tetraedro tiene asociado un vector de 3D (velocidad
    // del viento en el centro del tetraedro)
    fprintf(f, "%zu %zu 0 3 0\n", malla->Num_nodos, malla->Num_tetra);
    //	 Graba las coordenadas de los nodos
    for(size_t i=0; i < malla->Num_nodos; i++) {
        fprintf(f, "%zu %8.6f %8.6f %8.6f\n",
                i, malla->nodo[i][0], malla->nodo[i][1], malla->nodo[i][2]);
    }
    // Graba los tetraedros
    // num. ref., material, tipo, vertices
    for(size_t i=0; i < malla->Num_tetra; i++) {
        fprintf(f, "%zu 1 tet %zu %zu %zu %zu\n",
                i, malla->tetra[i][0], malla->tetra[i][1],
                malla->tetra[i][2], malla->tetra[i][3]);
    }

    // Para los tetraedros: num.comp., tamano comp1, tamano comp2, ...
    fprintf(f, "1 3\n");
    fprintf(f, "V, m/s\n");
    // Graba los datos de los tetraedros: vector velocidad y modulo
    for(size_t i=0; i < malla->Num_tetra; i++) {
        fprintf(f, "%zu %14.10f %14.10f %14.10f\n", i,
                malla->vini[i][0], malla->vini[i][1], malla->vini[i][2]);
    }

    fclose(f);
}


void GrabarModuloVientoAVS(const char *fich, const struct T_Malla *malla)
{
    FILE *f;

    if ( (f=fopen(fich,"w")) == NULL){
        fprintf(stderr, "No se pudo abrir el fichero");
        fprintf(stderr, "Nombre del fichero %s\n", fich);
        exit(EXIT_FAILURE);
    }
    // Graba el numero de nodos, numero de tetraedros, componentes de
    // datos de los nodos, componentes de datos de los tetraedros.
    // Cada tetraedro tiene asociado un escalar (modulo de la velocidad
    // del viento en el centro del tetraedro) y un vector de 3D (velocidad
    // del viento en el centro del tetraedro)
    fprintf(f, "%zu %zu 0 1 0\n", malla->Num_nodos, malla->Num_tetra);
    // Graba las coordenadas de los nodos
    for(size_t i=0; i < malla->Num_nodos; i++) {
        fprintf(f, "%zu %8.6f %8.6f %8.6f\n",
                i, malla->nodo[i][0], malla->nodo[i][1], malla->nodo[i][2]);
    }
    // Graba los tetraedros
    // num. ref., material, tipo, vertices
    for(size_t i=0; i < malla->Num_tetra; i++) {
        fprintf(f, "%zu 1 tet %zu %zu %zu %zu\n",
                i, malla->tetra[i][0], malla->tetra[i][1],
                malla->tetra[i][2], malla->tetra[i][3]);
    }

    // Para los tetraedros: num.comp., tamano comp1, tamano comp2, ...
    fprintf(f, "1 1\n");
    fprintf(f, "V, m/s\n");
    // Graba los datos de los tetraedros: vector velocidad y modulo
    // Buscamos modulo maximo y minimo

    DEBUG_INFO(
                double mod = sqrt((malla->wind[0][0]*malla->wind[0][0]) +
            (malla->wind[0][1]*malla->wind[0][1]) +
            (malla->wind[0][2]*malla->wind[0][2]));
    double mod_max = mod;
    double mod_min = mod;
    );

    for(size_t i=0; i < malla->Num_tetra; i++) {
        double modulo = sqrt((malla->wind[i][0]*malla->wind[i][0]) +
                (malla->wind[i][1]*malla->wind[i][1]) +
                (malla->wind[i][2]*malla->wind[i][2]));
        fprintf(f, "%zu %14.10f\n", i,
                modulo);

        DEBUG_INFO(
                    mod_max = fmax(mod_max,modulo);
                mod_min = fmin(mod_min,modulo);
        );

    }

    DEBUG_INFO(
                // Escribimos por pantalla mod_max y mod_min
                fprintf(stdout, "Modulo maximo = %f\nModulo minimo = %f\n", mod_max,
                        mod_min);
            fflush(stdout);
    );


    fclose(f);
}

double VolTetra(const double* x1, const double* y1, const double* z1,
                const double* x2, const double* y2, const double* z2,
                const double* x3, const double* y3, const double* z3,
                const double* x4, const double* y4, const double* z4
                )
{
    double A[3][3];

    A[0][0] = (*x2)-(*x1);   A[0][1] = (*y2)-(*y1);   A[0][2] = (*z2)-(*z1);
    A[1][0] = (*x3)-(*x1);   A[1][1] = (*y3)-(*y1);   A[1][2] = (*z3)-(*z1);
    A[2][0] = (*x4)-(*x1);   A[2][1] = (*y4)-(*y1);   A[2][2] = (*z4)-(*z1);

    return(0.1666666666666666 * djac(A));
}


int
BuscaPuntoEnMalla(const struct T_Malla  *malla,
                  const double *X, const double *Y, const double *Z)
{
    double vgrande, K;
    int tetra = -1;
    double dist_centetra;

    // Para cada tetraedro de la malla
    for(size_t i = 0; i < malla->Num_tetra; i++){
        // calcula el volumen del tetraedro
        vgrande = fabs(VolTetra (&malla->nodo[malla->tetra[i][0]][0],
                &malla->nodo[malla->tetra[i][0]][1],
                &malla->nodo[malla->tetra[i][0]][2],

                &malla->nodo[malla->tetra[i][1]][0],
                &malla->nodo[malla->tetra[i][1]][1],
                &malla->nodo[malla->tetra[i][1]][2],

                &malla->nodo[malla->tetra[i][2]][0],
                &malla->nodo[malla->tetra[i][2]][1],
                &malla->nodo[malla->tetra[i][2]][2],

                &malla->nodo[malla->tetra[i][3]][0],
                &malla->nodo[malla->tetra[i][3]][1],
                &malla->nodo[malla->tetra[i][3]][2]
                )
                );

        // Calcula los volumenes de cada tetraedro formado por el punto (X,Y,Z)
        // y tres de los nodos del tetraedro. Si la suma de los valores
        // absolutos de los  4 volumenes asi calculados es igual (dentro
        // de un margen) al del tetraedro considerado, es que el punto
        // esta en el interior
        double aux = 0.0;


        aux += fabs(VolTetra (X, Y, Z,

                              &malla->nodo[malla->tetra[i][1]][0],
                &malla->nodo[malla->tetra[i][1]][1],
                &malla->nodo[malla->tetra[i][1]][2],

                &malla->nodo[malla->tetra[i][2]][0],
                &malla->nodo[malla->tetra[i][2]][1],
                &malla->nodo[malla->tetra[i][2]][2],

                &malla->nodo[malla->tetra[i][3]][0],
                &malla->nodo[malla->tetra[i][3]][1],
                &malla->nodo[malla->tetra[i][3]][2]
                )
                );
        if (greatThan(&aux,&vgrande)){
            continue;
        }

        aux += fabs(VolTetra (&malla->nodo[malla->tetra[i][0]][0],
                &malla->nodo[malla->tetra[i][0]][1],
                &malla->nodo[malla->tetra[i][0]][2],

                X, Y, Z,

                &malla->nodo[malla->tetra[i][2]][0],
                &malla->nodo[malla->tetra[i][2]][1],
                &malla->nodo[malla->tetra[i][2]][2],

                &malla->nodo[malla->tetra[i][3]][0],
                &malla->nodo[malla->tetra[i][3]][1],
                &malla->nodo[malla->tetra[i][3]][2]
                )
                );
        if (greatThan(&aux,&vgrande)){
            continue;
        }

        aux += fabs(VolTetra (&malla->nodo[malla->tetra[i][0]][0],
                &malla->nodo[malla->tetra[i][0]][1],
                &malla->nodo[malla->tetra[i][0]][2],

                &malla->nodo[malla->tetra[i][1]][0],
                &malla->nodo[malla->tetra[i][1]][1],
                &malla->nodo[malla->tetra[i][1]][2],

                X, Y, Z,

                &malla->nodo[malla->tetra[i][3]][0],
                &malla->nodo[malla->tetra[i][3]][1],
                &malla->nodo[malla->tetra[i][3]][2]
                )
                );

        if (greatThan(&aux,&vgrande)){
            continue;
        }
        aux += fabs(VolTetra (&malla->nodo[malla->tetra[i][0]][0],
                &malla->nodo[malla->tetra[i][0]][1],
                &malla->nodo[malla->tetra[i][0]][2],

                &malla->nodo[malla->tetra[i][1]][0],
                &malla->nodo[malla->tetra[i][1]][1],
                &malla->nodo[malla->tetra[i][1]][2],

                &malla->nodo[malla->tetra[i][2]][0],
                &malla->nodo[malla->tetra[i][2]][1],
                &malla->nodo[malla->tetra[i][2]][2],

                X, Y, Z
                )
                );
        // si la suma de los 4 volumenes es aproximadamente igual
        // al volumen del tetraedro en cuestion, es que el punto
        // esta en el interior
        K = fabs(aux - vgrande);
        if (equal(&aux,&vgrande)) {
            if (tetra < 0) {
                tetra = i;
                dist_centetra = K;
            } else {
                fprintf(stderr,
                        "BuscaPuntoEnMalla: el punto (%f, %f, %f) ya se habia \
                        localizado en el tetraedro %d\ny se ha vuelto a localizar en el %zu\n",
                        *X, *Y, *Z, tetra, i);
                fflush(stderr);
                // Me quedo con la menor de las diferencias de volumen
                if (K < dist_centetra) {
                    tetra = i;
                    dist_centetra = K;
                    fprintf(stderr, "\tLa diferencia de volumen es menor en el \
                            tetraedro %zu,\n\tasi que cambia de tetraedro\n", i);
                }
            }
        }
    }

    return(tetra);
}



double DistanciaEntreNodos(const struct T_Malla *malla, const size_t nodo1, const size_t nodo2)
{
    double d = 0.0;

    for (size_t i=0; i<3; i++) {
        double s = malla->nodo[nodo1][i] - malla->nodo[nodo2][i];
        s = s*s;
        d += s;
    }
    d = sqrt(d);
    return(d);
}
