#ifndef CARAS_EXT_H
#define CARAS_EXT_H

#include "lectura.h"
/**
 * @file caras_ext.h
 * @brief Módulo para buscar los números de condición de los nodos frontera de una malla
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @brief Función que busca los números de condición de la malla y se crea la estructura terreno
 * @param[inout] malla Malla a la que buscar números de condición, se modifican los componentes condición y terreno
 */

void
CorreccionNumerosReferencia(struct T_Malla *malla);

/**
 * @brief Función que dados 2 tetraedros mira si tiene caras comunes y las marca como interiores
 * @param[in] tet Número de tetraedo 1 a comparar
 * @param[in] cara Cara a comparar del tetraedro 1
 * @param[in] nods_cara Nodos que componen la cara 1
 * @param[in] tet2  Número de tetraedo 2 a comparar
 * @param[in] cara2 Cara a comparar del tetraedro 2
 * @param[in] nods_cara2 Nodos que componen la cara 2
 * @param[out] tetraedros_caras Si es una cara interior modifica su valor a la marca de interior (1)
 */
void BuscarCaraInterior (const size_t tet, const size_t cara, const size_t *nods_cara, const size_t tet2,
        const size_t cara2, const size_t *nods_cara2, size_t **tetraedros_caras);

/**
 * @brief Función que dada un número de cara de un tetraedro devuelve los indices de sus tres nodos
 * @param[in] no_cara Número de la cara a encontrar
 * @param[out] cara1_id,cara2_id,cara3_id Índice 1, 2 y 3 de los nodos de la cara
 */
void DevuelveCaras (size_t no_cara, size_t *cara1_id, size_t *cara2_id, size_t *cara3_id);

/**
 * @brief Función que dado un tetraedro y un número de cara devuelve los índices de los 3 nodos de la cara
 * @param[in] tetra Conectividades de los tetraedros
 * @param[in] tet Número de tetraedro a consultar
 * @param[in] cara Número de cara que se requiere
 * @param[out] nodos_cara Vector con los índices de los 3 nodos de la cara
 */
void DevuelveNodosCara(const size_t **tetra, const size_t tet, const size_t cara, size_t *nodos_cara);

/**
 * @brief Función que dado un escalar y un vector comprueba si alguno de los componentes del vector es igual al escalar
 * @param[in] scalar Valor del escalar a comparar
 * @param[in] vector Vector al que comparar
 * @param[in] size Tamaño del vector
 * @return Verdadero si el escalar se encuentra en el vector, falso en caso contrario
 */
bool any(const size_t scalar, const size_t* vector, const size_t size);

/**
 * @brief Función que sirve para localizar caras con ciertos números de referencia
 * @param[in] tet Tetraedro al que pertenece la cara
 * @param[in] nodo1,nodo2,nodo3 Nodo 1, 2 y 3 de la cara
 * @param[in] ref1,ref2,ref3 referencia 1, 2 y 3 a las que comparar
 * @param[in] ref_cara Valor de referencia de la cara
 * @param[in] cara Números de referencia de la cara
 * @param[out] nodos_pared Lista de nodos que pertenecen a una cara
 * @param[out] ref_caras Lista de referencias de caras para cada tetraedro
 */
void LocalizarCara(const size_t tet, const size_t nodo1, const size_t nodo2, const size_t nodo3, const size_t ref1,
        const size_t ref2, const size_t ref3, const size_t ref_cara, const size_t *cara, size_t **nodos_pared,
        size_t *ref_caras);

#endif //CARAS_EXT_H
