#ifndef COMPARISONS_H
#define COMPARISONS_H

#include <math.h>
#include <stdbool.h>

/**
 * @file comparisons.h
 * @brief Fichero con evaluaciones de comparación con tolerancia
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @brief Valor de la tolerancia usada para las comparaciones
 */
static const double comparisons_tol = 1.0e-12;

/**
 * @brief Función que devuelve si un número es igual a zero
 * @param[in] x0 Número a comparar
 * @return Verdadero si el número es igual a zero, falso en caso contrario
 */
static inline bool
equalToZero(const double *x0)
{
    return fabs((*x0)) < comparisons_tol;
}

/**
 * @brief Función que devuelve si un número es menor que zero
 * @param[in] x0 Número a comparar
 * @return Verdadero si el número es menor que zero, falso en caso contrario
 */
static inline bool
lessThanZero(const double *x0)
{
  if (equalToZero(x0))
    {
      return false;
    }
    return -(*x0) > comparisons_tol;
}

/**
 * @brief Función que comprueba si un número es mayor que zero
 * @param[in] x0 Número a comparar
 * @return Verdadero si el número es mayor a zero, falso en caso contrario
 */
static inline bool
greatThanZero(const double *x0)
{
  if (equalToZero(x0))
    {
      return false;
    }
    return (*x0) > comparisons_tol;
}

/**
 * @brief Función que comprueba si un número es mayor o igual a zero
 * @param[in] x0 Número a comparar
 * @return Verdadero si es mayor o igual a zero, falso en caso contrario
 */
static inline bool
greatThanOrEqualToZero(const double *x0)
{
  return(greatThanZero(x0) || equalToZero(x0));
}

/**
 * @brief Función que comprueba si un número es menor o igual a zero
 * @param[in] x0 Número a comparar
 * @return Verdadero si el número es menor o igual a zero, falso en caso contrario
 */
static inline bool
lessThanOrEqualToZero(const double *x0)
{
  return(lessThanZero(x0) || equalToZero(x0));
}

/**
 * @brief Función que comprueba si dos números son iguales
 * @param[in] x0 Primer número a comparar
 * @param[in] x1 Segundo número a comparar
 * @return Verdadero si los dos números son iguales, falso en caso contrario
 */
static inline bool
equal(const double *x0, const double *x1)
{
    return fabs((*x0)-(*x1)) < comparisons_tol;
}

/**
 * @brief Función que comprueba si un número es menor a otro
 * @param[in] x0 Primer número a comparar
 * @param[in] x1 Segundo número a comparar
 * @return Verdadero si x0 es menor que x1, falso en caso contrario
 */
static inline bool
lessThan(const double *x0, const double *x1)
{
  if (equal(x0,x1))
    {
      return false;
    }
    return (*x1)-(*x0) > comparisons_tol;
}

/**
 * @brief Función que comprueba si un número es mayor a otro
 * @param[in] x0 Primer número a comparar
 * @param[in] x1 Segundo número a comparar
 * @return Verdadero si x0 es mayor que x1, falso en caso contrario
 */
static inline bool
greatThan(const double *x0, const double *x1)
{
  if (equal(x0,x1))
    {
      return false;
    }
    return (*x0)-(*x1) > comparisons_tol;
}

/**
 * @brief Función que comprueba si un número es mayor o igual a otro
 * @param[in] x0 Primer número a comparar
 * @param[in] x1 Segundo número a comparar
 * @return Verdadero si x0 es mayor o igual a x1, falso en caso contrario
 */
static inline bool
greatThanOrEqual(const double *x0, const double *x1)
{
  return(greatThan(x0,x1) || equal(x0,x1));
}

/**
 * @brief Función que comprueba si un número es menor o igual a otro
 * @param[in] x0 Primer número a comparar
 * @param[in] x1 Segundo número o comparar
 * @return Verdadero si x0 es menor o igual a x1, falso en caso contrario
 */
static inline bool
lessThanOrEqual(const double *x0, const double *x1)
{
  return(lessThan(x0,x1) || equal(x0,x1));
}



#endif // COMPARISONS_H
