#ifndef ENSAMBLAJE_H
#define ENSAMBLAJE_H

#include "grad.h"
#include "lectura.h"
/**
 * @file ensamblaje.h
 *
 * @brief Estructuras para la creacion de la matriz morse a partir de las
 * matrices elementales.
 *
 * Se crea un vector de listas encadenadas
 * que se va rellenando con las matrices elementales, para despues
 * construir la matriz morse a partir de el.
 *
 * Contiene tambien los prototipos de las funciones que pueden
 * ser invocadas desde otros modulos
 *
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @brief Estructura de nodo para las matrices escasas
 */
struct T_Nodo{
    /**
   * @brief columna de la matriz completa
   */
    size_t col;
    /**
   * @brief valor del elemento
   */
    double valor;
    /**
   * @brief Puntero para la lista enlazada
   */
    struct T_Nodo *ptr;
};

/**
 * @brief Estructura de vector de nodos para las matrices escasas
 */

struct T_Vector_Nodos{
    /**
   * @brief Numero de nodos de la lista encadenada
   */
    size_t count;
    /**
   * @brief Valor del término independiente
   */
    double xp;
    /**
   * @brief Puntero para la lista enlazada
   */
    struct T_Nodo *ptr;
};

/**
 * @brief Estructura de matriz elemental
 */
struct T_MElemental{
    /**
   * @brief Tamano del vector = numero de incognitas
   */
    size_t dim;
    /**
   * @brief vector de traduccion (eltos dirichlet no se ensamblan)
   */
    int *indice;
    /**
   * @brief Valores de la matriz
   */
    double **valor;
};

/**
 * @brief Función que genera una estructura T_Vector_Nodos vacia de dimensión n
 * @param[in] n Dimensión de la estructura
 * @return Puntero a la estructura generada en la función
 */
struct T_Vector_Nodos *
        CrearVectorDeListas(const size_t n);

/**
 * @brief Función que libera la memoria usada por el vector de listas encadenadas
 *        que sirve para ensamblar la matriz morse
 * @param[inout] matriz Matriz a liberar
 * @param[in] n Dimensión de la matriz
 */
void
LiberarVectorDeListas(struct T_Vector_Nodos *matriz, const size_t n);


/**
 * @brief Funcion que reserva memoria para una matriz elemental de dimension dim
 * @param[in] dim Dimensión de la matriz elemental a crear
 * @return Puntero a la matriz creada. Si no puede reservar memoria aborta el programa
 */
struct T_MElemental *
        CrearMatrizElemental(const size_t dim);


/**
 * @brief Funcion que libera la memoria usada por una matriz elemental
 * @param[inout] eltal1 puntero a la matriz elemental que se desea destruir
 */
void
DestruirMatrizElemental(struct T_MElemental *eltal1);

/**
 * @brief Función que dada una matriz elemental la ensambla en la matriz
 * @param[inout] Matriz Matriz en la que se ensambla la matriz elemental
 * @param[in] eltal Matriz elemental a incrustar
 * @param[in] malla Malla para conocer el tipo de elemento
 * @param[in] xp Valor del término independiente
 */
void 
EnsamblarMatrizElemental(struct T_Vector_Nodos *Matriz, const struct T_MElemental *eltal,
                         const struct T_Malla *malla, const double *xp);


/**
 * @brief Función que crea la matriz morse del sistema reducido,
 *        el vector de terminos independientes del sist. reducido y
 *        el vector de reensamblaje
 * @param[in] Matriz Matriz ensamblada del sistema
 * @param[in] dim Dimensión del sistema
 * @param[out] morse Matriz morse del sistema
 * @param[out] x Vector de terminos independientes del sistema
 * @param[out] P Vector de reensamblaje
 * @return La dimension del vector de reensamblaje
 */
size_t
ConstruirMatrizMorse(const struct T_Vector_Nodos *Matriz, const size_t dim,
                     struct matrizmorse *morse, double **x, size_t **P);


/**
 * @brief Funcion que genera el vector solucion completo
 *        (incluidos nodos tipo Dirichlet) a partir del vector solucion parcial
 *        y el vector de reensamblaje.
 * @param[in] dim Dimension del sistema completo (num. de nodos de la malla)
 * @param[in] x Vector solucion incompleto (sin nodos Dirichlet)
 * @param[inout] r Vector de reensamblaje
 * @return Puntero al vector solucion completo del sistema
 */
double *
Reensamblar(const size_t dim, const double *x, size_t *r);

/**
 * @brief Función que genera y resuelve el sistema de equaciones reducido
 * @param[out] Matriz Matriz morse en la que resolver el sistema
 * @param[in] dim Dimensión de la matriz
 * @param[in] metodo Método de resolución del sistema (GradienteConjugado o BICGSTAB)
 * @param[in] precond Método de precondicionador (PrecondDiag, PrecondDiagOptimo, PrecondILU o PrecondSSOR_NS)
 * @param[in] w Parámetro del precondicionador
 * @return Vector resultado
 */
double *
ResolverSistema(struct T_Vector_Nodos *Matriz, const size_t dim, const int metodo,
                const int precond, const double w);

/**
 * @brief Función que inserta un nodo en el vector de nodos
 * @param[inout] Matriz Matriz en la que insertar el nodo
 * @param[in] fila fila de la matriz a la que pertenece el nodo
 * @param[in] col columna de la matriz a la que pertenece el nodo
 * @param[in] valor valor del nodo a insertar
 * @param[in] xp valor del termino independiente
 */
void InsertarNodo(struct T_Vector_Nodos *Matriz, const size_t fila, const size_t col, const double valor,
                  const double xp);

/**
 * @brief Función que dado un número de referencia devuelve el tipo de condición de contorno (Dirichlet, Newman) de la cara
 * @param[in] ref Número de referencia de la cara
 * @return Tipo de condición de contorno
 */
int Condicion(const size_t ref);

#endif // ENSAMBLAJE_H
