#include "grad.h"

#include "comparisons.h"
#include "debugInfo.h"

#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

/**
 * @file grad.c
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

void DestruirMorse(struct matrizmorse *A)
{
    free(A->valor);
    free(A->columna);
    free(A->fila);
    free(A);
}


void ProductoMorse(const struct matrizmorse *M, const double *v, double *vres)
{
    const size_t pfilaEnd = M->dimension;

    for (size_t pfila = 0; pfila < pfilaEnd; pfila++) {
        double suma = 0.0;
        const size_t nfEnd = M->fila[pfila+1];
        for (size_t nf = M->fila[pfila]; nf < nfEnd; nf++) {
            suma = suma + (M->valor[nf] * v[M->columna[nf]]);
        }
        vres[pfila] = suma;
    }
}

double ProductoVectores(const double *a,const double *b,const size_t dim)
{
    double s=0.0;

    for (size_t i=0; i<dim; i++) {
        s += a[i] * b[i];
    }

    return(s);

} // ProductoVectores


// Funcion que efectua el calculo de la norma 2 de un vector

double Norma2(const double *a, const size_t dim)
{

    double sum = 0;

    for (size_t i=0; i<dim; i++) {
        sum = sum + (a[i]*a[i]);
    }

    return(sqrt(sum));

} // Norma2


void DiagonalMorse(const struct matrizmorse *M, double *z)
{

    DEBUG_INFO(
                fprintf(stderr, "DEBUG: DiagonalMorse: Entro en DiagonalMorse\n");
            );


    const size_t pfilaEnd = M->dimension;

    for (size_t pfila = 0; pfila < pfilaEnd; pfila++) {
        z[pfila] = 0.0;
        const size_t nfEnd = M->fila[pfila+1];
        for (size_t nf = M->fila[pfila]; nf < nfEnd; nf++) {
            if (pfila == M->columna[nf])
                z[pfila] = M->valor[nf];
        }
    }



    DEBUG_INFO(
                // Busca un 0 en la diagonal de M

                for (size_t pfila=0;  pfila < pfilaEnd; pfila++) {
        double z_fila = fabs(z[pfila]);
        if (equalToZero(&z_fila)) {
            fprintf(stderr, "DEBUG: PosDiagonalMorse: Hay un 0 en la diagonal de M\n");
            fprintf(stderr, "DEBUG: PosDiagonalMorse: M[%zu]=%g\n", pfila,
                    z[pfila]);
            exit(EXIT_FAILURE);
        }
    }
    fprintf(stderr, "DEBUG: DiagonalMorse: Salgo de DiagonalMorse\n");
    );


} // DiagonalMorse


// Modifica z para que contenga las posiciones de la diagonal de M
void PosDiagonalMorse(const struct matrizmorse *M, size_t *z)
{
    const size_t pfilaEnd = M->dimension;
    for (size_t pfila = 0; pfila < pfilaEnd; pfila++) {
        z[pfila] = -1;
        const size_t nfEnd = M->fila[pfila+1];
        for (size_t nf = M->fila[pfila]; nf < nfEnd; nf++) {
            if (pfila == M->columna[nf])
                z[pfila] = nf;
        }
    }

    DEBUG_INFO(
                // Busca un inf en la diagonal de M
                for (size_t pfila=0;  pfila < pfilaEnd; pfila++) {
        if ( !isfinite(fabs(M->valor[z[pfila]])) ) {
            fprintf(stderr,
                    "DEBUG: PosDiagonalMorse: inf en la diagonal (fila %zu) (%f)\n",
                    pfila, M->valor[z[pfila]]);
            exit(EXIT_FAILURE);
        }
    }
    );

} // PosDiagonalMorse



size_t BuscaColumna(const struct matrizmorse *A, const size_t col, double *vec)
{
    size_t k = 0;
    const size_t filaEnd = A->dimension;
    for (size_t fila=0; fila < filaEnd; fila++){
        const size_t iEnd = A->fila[fila+1];
        for (size_t i = A->fila[fila]; i < iEnd; i++) {
            if (A->columna[i] == col){
                vec[k] = A->valor[i];
                k++;
                break;
            }
        }
    }
    return(k);
}

double * 
PrecondDiag(const struct matrizmorse *A, const size_t *posdiag_A, const double *w, size_t *posdiag_W )
{
    double *p;


    DEBUG_INFO(
                fprintf(stderr, "Entro en PrecondDiag\n");
            );

    if ( (p=(double *)malloc(A->nonulos * sizeof(double))) == NULL){
        fprintf(stderr, "PrecondDiag: No se ha podido alocatar p\n");
        exit(EXIT_FAILURE);
    }
    DiagonalMorse(A, p);


    for(size_t i=0; i< A->dimension; i++)
    {
        posdiag_W[i] = i;
    }

    return p;
} // PrecondDiagonal


/**
 * Sólo usa el parametro A; los demas se ignoran
 */
double * 
PrecondDiagOptimo(const struct matrizmorse *A, const size_t *posdiag_A, const double *w, size_t *posdiag_W)
{
    double *p, *col;
    size_t *posdiag;

    if ( (p=(double *)malloc(A->nonulos * sizeof(double))) == NULL){
        fprintf(stderr, "PrecondDiagOptimo: No se ha podido alocatar p\n");
        exit(EXIT_FAILURE);
    }
    if ( (posdiag=(size_t *)malloc(A->dimension * sizeof(size_t))) == NULL){
        fprintf(stderr, "PrecondDiagOptimo: No se ha podido alocatar posdiag\n");
        exit(EXIT_FAILURE);
    }
    if ( (col=(double *)malloc(A->dimension * sizeof(double))) == NULL){
        fprintf(stderr, "PrecondDiagOptimo: No se ha podido alocatar col\n");
        exit(EXIT_FAILURE);
    }
    PosDiagonalMorse(A, posdiag);
    for(size_t i=0; i < A->dimension; i++){
        size_t n = BuscaColumna(A, i, col);
        double n2 = Norma2(col, n);
        p[i] = (n2 * n2)/ A->valor[posdiag[i]];
    }

    for(size_t i=0; i< A->dimension; i++)
        posdiag_W[i] = i;
    free(col); free(posdiag);
    return p;
} // PrecondDiagonalOptimo

double * PrecondSSOR_NS(const struct matrizmorse *A, const size_t *posdiag_A, const double *w,
                        size_t *posdiag_W){
    double *p;

    if ( (p=(double *)malloc(A->nonulos * sizeof(double))) == NULL){
        fprintf(stderr, "PrecondSSOR_NS: No se ha podido alocatar p\n");
        exit(EXIT_FAILURE);
    }

    for (size_t t=0; t < A->dimension; t++){
        for (size_t i = A->fila[t]; i < A->fila[t+1]; i++){
            if (A->columna[i] < t){ // triangular inferior
                p[i] = (*w) * A->valor[i] / A->valor[posdiag_A[A->columna[i]]];
            }
            else{
                if (A->columna[i] == t) { // diagonal
                    p[i] = A->valor[i] / (*w) / (2.0 - (*w));
                    posdiag_W[t] = i;
                } else { // triangular superior
                    p[i] =  (A->valor[i] / (2.0 -(*w)));
                }
            }
        }
    }
    return p;
} // PrecondSSOR_NS

size_t BuscaEnFila(const struct matrizmorse *A, const size_t f, const size_t c)
{

    size_t sitio = A->dimension+1;
    size_t i = A->fila[f];

    while (i < A->fila[f+1]){
        if (A->columna[i] == c){
            sitio = i;
            break;
        }
        i++;
    }

    return(sitio);
} // BuscaEnFila


double * PrecondILU(const struct matrizmorse *A, const size_t *posdiag_A, const double *w, size_t *posdiag_W)
{
    //    int p, q, x, d;
    double *W, suma;


    W = (double *)malloc(A->nonulos * sizeof(double));

    if (W == NULL) {
        fprintf(stderr, "PrecondILU: No se ha podido alocatar W\n");
        exit(EXIT_FAILURE);
    }

    // Iguala W a A, y de paso ya tenemos la primera fila de la U calculada
    for (size_t i=0; i < A->nonulos; i++){
        W[i] = A->valor[i];
    }
    // Calculo de la primera columna de la L
    for (size_t i=1; i < A->dimension; i++){
        size_t j = A->fila[i];
        if (A->columna[j] == 0){
            W[j] = A->valor[j] / W[posdiag_A[0]];
        }
    }


    for (size_t f=1; f < A->dimension; f++){
        // Calculo de la U
        size_t x = posdiag_A[f];
        while (x < A->fila[f+1]) {
            size_t p = A->columna[x];
            size_t j = f;
            size_t q = A->fila[j];
            suma = 0.0;
            while (q < posdiag_A[j]){
                size_t d = A->columna[q];
                // k = BuscaEnFila(A, d, p);
                // Busca en la fila d-esima el elemento con columna p; si existe
                // ha de estar en la triangular superior o en la diagonal
                size_t k = posdiag_A[d];
                while ((k < A->fila[d+1]) && (A->columna[k] != p))
                    k++;
                // Si ese elto existe, su posicion es k
                if (k < A->fila[d+1])
                    suma = suma + W[k] * W[q];
                q++;
            }
            W[x] = A->valor[x] - suma;
            x++;
        }

        // Calculo de la L
        x = f+1;
        while ( x < A->dimension){
            // Busca en la fila x-esima el elemento con columna f; si existe
            // ha de estar en la triangular inferior
            size_t k = A->fila[x];
            while ((k < posdiag_A[x]) && (A->columna[k] != f))
                k++;
            // Si ese elto existe, su posicion es k
            if (k < posdiag_A[x]){
                suma = 0.0;
                size_t j=A->fila[x];
                while (j < k) {
                    size_t p = A->columna[j];
                    // Busca en la fila p-esima el elemento con columna f; si existe
                    // ha de estar en la triangular superior
                    size_t h = posdiag_A[p] + 1;
                    while ((h < A->fila[p]) && (A->columna[h] != f))
                        h++;
                    // Si ese elto existe, su posicion es h
                    if (h < A->fila[p])
                        suma = suma + W[j] * W[h];
                    j++;
                }
                W[k] = (A->valor[k] - suma) / W[posdiag_A[f]];
            }
            x++;
        }
    }

    for (size_t i=0; i< A->dimension; i++)
        posdiag_W[i] = posdiag_A[i];

    return W;
} // PrecondILU

void RemonteDiag(const double *r, const double *W, const size_t *posdiag_W, const struct matrizmorse *A, double *y){

    const size_t iEnd = A->dimension;
    for (size_t i=0; i < iEnd; i++){
        y[i] = r[i]  / W[posdiag_W[i]];
    }
}


void RemonteIdent(const double *r, const double *W, const size_t *posdiag_W, const struct matrizmorse *A, double *y){
    memmove((void *) y, (void *) r, (size_t)(A->dimension * sizeof(double)) );
}

void RemonteLU(const double *r, const double *W, const size_t *posdiag_W, const struct matrizmorse *A, double *y)
{

    for (size_t i=0; i<A->dimension; i++)
    {
        y[i] = r[i];
    }

    for (size_t i=1; i <= A->dimension - 1; i++){
        double suma = 0.0;
        size_t j = A->fila[i];
        size_t k = A->fila[i+1];
        while ((A->columna[j] < i) && (j < k)) {
            suma = suma + (y[A->columna[j]] * W[j]);
            j++;
        }
        y[i] = y[i] - suma;
    }

    y[A->dimension - 1] /= W[posdiag_W[A->dimension - 1]];
    for (size_t i=A->dimension - 1; i > 0; i--){
        double suma = 0.0;
        size_t j = A->fila[i] - 1;
        size_t k = A->fila[i-1];
        while ((A->columna[j] > i-1) && (j >= k)){
            suma = suma + (W[j] * y[A->columna[j]]);
            j--;
        }
        y[i-1] = (y[i-1] - suma) / W[posdiag_W[i-1]];
    }
}


bool GradienteConjugado(const struct matrizmorse *A, const double *b, double *solucion, const double *u0,
                        const size_t numiter, const double toler, const double *W,
                        const size_t *posdiag_W, const TFuncRemonte remonte){

    size_t  niter=0;
    double alfa, beta;
    double alfa_num, alfa_den, beta_num, beta_den, Norm_r0, Norm_r, condicion;
    double LaNorma;

    double* prod = (double*) malloc(sizeof(double) * A->dimension);
    double* p = (double*) malloc(sizeof(double) * A->dimension);
    double* r = (double*) malloc(sizeof(double) * A->dimension);
    double* z = (double*) malloc(sizeof(double) * A->dimension);
    double* mi_r = (double*) malloc(sizeof(double) * A->dimension);
    double* mi_prod = (double*) malloc(sizeof(double) * A->dimension);

    for(size_t i=0; i < A->dimension; i++) {
        solucion[i] = u0[i];
    }
    // Comprobacion de que Vinicial != solucion

    ProductoMorse(A, u0, prod);
    for(size_t i=0; i < A->dimension; i++){
        r[i] = b[i] - prod[i];
    }

    Norm_r0 = Norma2(r, A->dimension);
    // La aproximacion inicial coincide con la solucion
    if (Norm_r0 < toler){
        fprintf(stdout, "La aproximacion inicial coincide con la solucion\n");
        return(true);
    }

    remonte(r, W, posdiag_W, A, z);
    for(size_t i=0; i < A->dimension; i++){
        p[i] = z[i];
    }

    alfa_num = ProductoVectores(r, z, A->dimension);
    Norm_r = Norm_r0;
    condicion = Norm_r / Norm_r0;

    do
    {
        while (( condicion > toler ) && (niter <= numiter) ){

            DEBUG_INFO(
                        fprintf(stderr,
                                "DEBUG: Grad: Iteracion %zu  condicion=%g tol=%g\n",
                                niter, condicion, toler);
                    );

            niter++;
            ProductoMorse(A, p, prod);
            alfa_den = ProductoVectores(p, prod, A->dimension);
            //La igualación tiene que ser exacta.
            //En casos donde alfa_den es muy pequeño (~1.0e-32) el sistema funciona
            if (alfa_den == 0.0){
                fprintf(stdout,"Grad: ERROR: Denominador de alfa es 0: %g\n", alfa_den);
                return(false);
            }
            assert(isfinite(alfa_den));
            alfa = alfa_num /alfa_den;
            assert(isfinite(alfa));
            for (size_t t=0; t< A->dimension; t++) {
                solucion[t] = solucion[t] + alfa * p[t];
                r[t] = r[t] - alfa * prod[t];
            }

            remonte(r, W, posdiag_W, A, z);

            beta_num = ProductoVectores(r, z, A->dimension);
            beta_den = alfa_num;
            beta = beta_num / beta_den;

            for(size_t t=0; t < A->dimension; t++)
                p[t] = z[t] + beta * p[t];

            alfa_num = beta_num;

            Norm_r = Norma2(r, A->dimension);
            condicion = Norm_r / Norm_r0;
        }

        // Comprueba que realmente Ax=b
        ProductoMorse(A, solucion, mi_prod);
        for(size_t i=0; i < A->dimension; i++){
            mi_r[i] = b[i] - mi_prod[i];
        }

        LaNorma = Norma2(mi_r, A->dimension);
        LaNorma = LaNorma /Norm_r0;

        DEBUG_INFO(
                    fprintf(stdout, "La precision requerida es %g\nLa precision real es %g\n",
                           toler, LaNorma);
                );

        if (LaNorma >= toler){
            DEBUG_INFO(
                        fprintf(stdout, "Saliendo del while, con precision real %g\n", LaNorma);
                    fprintf(stdout, "Numero de iteraciones max: %zu\trealizadas: %zu\n",
                           numiter, niter);
            );

            condicion = toler + 1;
        }
    }
    while ((toler < LaNorma) && (niter <= numiter));
    if (toler < LaNorma)
    {
        fprintf(stderr, "Se salio del while externo, pero la precision real es %g\n",
               LaNorma);
        fflush(NULL);
        exit(EXIT_FAILURE);
    }

    free(prod); free(p); free(r); free(z); free(mi_r); free(mi_prod);
    if (niter >= numiter)
    {
        fprintf(stderr, "Grad: ERROR: no se resolvio el sistema\n"); fflush(NULL);
        exit(EXIT_FAILURE);
    }
    return (niter <= numiter);
} // Grad

bool BICGSTAB(const struct matrizmorse *A, const double *b, double *solucion, const double *u0,
              const size_t numiter, const double toler, const double *W, const size_t *posdiag_W,
              const TFuncRemonte remonte)
{

    size_t niter=0;
    double alfa, ro, ro_1, beta, w, w_den, Norm_r, Norm_r0, condicion;
    double LaNorma;

    double* prod = (double*) malloc(sizeof(double) * A->dimension);
    double* p = (double*) malloc(sizeof(double) * A->dimension);
    double* r = (double*) malloc(sizeof(double) * A->dimension);
    double* r0 = (double*) malloc(sizeof(double) * A->dimension);
    double* y = (double*) malloc(sizeof(double) * A->dimension);
    double* t = (double*) malloc(sizeof(double) * A->dimension);
    double* s = (double*) malloc(sizeof(double) * A->dimension);
    double* u = (double*) malloc(sizeof(double) * A->dimension);
    double* v = (double*) malloc(sizeof(double) * A->dimension);
    double* z = (double*) malloc(sizeof(double) * A->dimension);
    double* mi_r = (double*) malloc(sizeof(double) * A->dimension);

    for(size_t i=0; i < A->dimension; i++){
        solucion[i] = u0[i];
    }

    ProductoMorse(A, u0, prod);
    for(size_t i=0; i < A->dimension; i++){

        DEBUG_INFO(
                    fprintf(stderr, "BICGSTAB: A*U0[%zu]=%f\n", i, prod[i]);
                );

        r[i] = b[i] - prod[i];
        r0[i] = r[i];
        p[i] = v[i] = 0.0;
    }
    ro = alfa = w = 1.0;

    Norm_r0 = Norma2(r, A->dimension);
    // La aproximacion inicial coincide con la solucion
    if (Norm_r0 < toler)
    {
        return(true);
    }

    Norm_r = Norm_r0;
    condicion = Norm_r / Norm_r0;

    do
    {
        while (( condicion > toler ) && (niter <= numiter) ){

            DEBUG_INFO(
                        fprintf(stderr,
                                "BICGSTAB: Iteracion %zu  condicion=%10.6f tol=%e\n",
                                niter, condicion, toler);
                    );

            niter++;

            remonte(r, W, posdiag_W, A, z);
            ro_1 = ro;
            ro = ProductoVectores(r0, z, A->dimension);
            beta = (ro / ro_1) * (alfa / w);
            for (size_t i=0; i< A->dimension; i++){
                p[i] = z[i] + beta * (p[i] - w * v[i]);
            }
            ProductoMorse(A, p, y);

            remonte(y, W, posdiag_W, A, v);
            alfa = ProductoVectores(r0, v, A->dimension);
            //La igualación tiene que ser exacta.
            //En casos donde alfa_den es muy pequeño (~1.0e-32) el sistema funciona
            if (alfa == 0.0){
                fprintf(stdout,"BICGSTAB: ERROR: Denominador de alfa es 0: %g\n", alfa);
                return(false);
            }
            alfa = ro / alfa;
            for (size_t i=0; i< A->dimension; i++){
                s[i] = r[i] - alfa * y[i];
                u[i] = z[i] - alfa * v[i];
            }
            ProductoMorse(A, u, t);
            w_den = ProductoVectores(t, t, A->dimension);
            w = ProductoVectores(t, s, A->dimension);
            w = w / w_den;

            for (size_t i=0; i< A->dimension; i++) {
                solucion[i] = solucion[i] + alfa * p[i] + w * u[i];
                r[i] = s[i] - w * t[i];
            }

            Norm_r = Norma2(r, A->dimension);
            condicion = Norm_r / Norm_r0;
        }
        // Comprueba que realmente Ax=b
        ProductoMorse(A, solucion, prod);
        for(size_t i=0; i < A->dimension; i++){
            mi_r[i] = b[i] - prod[i];
        }

        LaNorma = Norma2(mi_r, A->dimension);
        LaNorma = LaNorma /Norm_r0;

        DEBUG_INFO(
                    fprintf(stdout, "BICGSTAB: La precision requerida es %g\nBICGSTAB: La precision real es %g\n",
                           toler, LaNorma);
                );

        if (LaNorma >= toler){

            DEBUG_INFO(
                        fprintf(stdout, "BICGSTAB: Saliendo del while, con precision real %g\n",
                               LaNorma);
                    fprintf(stdout, "BICGSTAB: Numero de iteraciones max: %zu\trealizadas: %zu\n",
                           numiter, niter);
            );

            condicion = toler + 1;
        }
    }
    while ((toler < LaNorma) && (niter <= numiter));
    if (toler < LaNorma)
    {
        fprintf(stderr, "Se salio del while externo, pero la precision real es %g\n",
               LaNorma);
        exit(EXIT_FAILURE);
    }

    free(prod); free(r); free(r0); free(p); free(y);
    free(s);  free(t); free(u);free(v);free(z);
    if (niter >= numiter)
    {
        fprintf(stderr, "BICGSTAB: ERROR: no se resolvio el sistema\n");
        exit(EXIT_FAILURE);
    }
    return (niter <= numiter);
} // BICGSTAB
