#ifndef GRAD_H
#define GRAD_H

#include <stdbool.h>
#include <stdio.h>
#include <time.h>

/**
 * @file grad.h
 * @brief Módulo donde se definen los métodos numéricos que trabajan sobre la matriz morse
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @brief Estructura para almacenar una matriz morse
 */
struct matrizmorse {
    /**
   * @brief Dimensión de la matriz
   */
  size_t dimension;
  /**
   * @brief Número de nodos no nulos
   */
  size_t nonulos;
  /**
   * @brief Puntero al vector de los valores de la matriz
   */
  double *valor;
  /**
   * @brief Puntero al número de elementos por columna
   */
  size_t *columna;
  /**
   * @brief Puntero al número de elementos por fila
   */
  size_t *fila;
};

/**
 * @brief DestruirMorse Función que libera la memoria para una estructura del tipo matrizmorse
 * @param[inout] A Estructura a destruir
 */
void DestruirMorse(struct matrizmorse *A);


/**
 * @brief Typedef que sirve para esconder un puntero a una función de precondicionador
 */
typedef double * (*TFuncPrecond) (const struct matrizmorse *A, const size_t *posdiag_A,
                  const double *w, size_t *posdiag_W);
/**
 * @brief Typedef que sirve para esconder un puntero a una función de Remonte
 */
typedef void (*TFuncRemonte) (const double *r, const double *W, const size_t *posdiag_W,
                  const struct matrizmorse *A, double *y);
/**
 * @brief Typedef que sirve para esconder un puntero a una función Solver
 */
typedef bool (*TFuncMetodo)(const struct matrizmorse *matriz, const double *b, double *solucion,
               const double *u0, const size_t numiter, const double toler, const double *W,
               const size_t *posdiag_W, const TFuncRemonte f);

/**
 * @brief Funcion que realiza el producto matriz x vector en almacenamiento morse
 * @param[in] M Matriz morse
 * @param[in] v Vector
 * @param[out] vres Vector resultante de la multiplicación
 */
void ProductoMorse(const struct matrizmorse *M, const double *v, double *vres);

/**
 * @brief Funcion que calcula los valores de la diagonal de una matriz morse
 * @param[in] M Matriz morse
 * @param[out] z vector con los valores de la diagonal
 */
void DiagonalMorse(const struct matrizmorse *M, double *z);

/**
 * @brief Función que calcula las posiciones de la diagonal de M
 * @param[in] M Matriz morse
 * @param[out] z vector con las posiciones de la diagonal de M
 */
void PosDiagonalMorse(const struct matrizmorse *M, size_t *z);

/**
 * @brief Función que devuelve la norma2 de un vector
 * @param[in] a Vector del que se calculará la norma
 * @param[in] dim dimensión del vector
 * @return Norma 2 del vector a
 */
double Norma2(const double *a, const size_t dim);

/**
 * @brief ProductoVectores Funcion que efectua el producto escalar a*b
 * @param[in] a Vector a
 * @param[in] b Vector b
 * @param[in] dim Dimensión de los vectores
 * @return Valor del producto escalar
 */
double ProductoVectores(const double *a, const double *b, const size_t dim);

/**
 * @brief Función que calcula el precondicionador diagonal de A
 * @param[in] A Matriz A del sistema
 * @param[in] posdiag_A Posiciones de la diagonal de la matriz A (no usado)
 * @param[in] w Valor del precondicionador (no usado)
 * @param[out] posdiag_W Posiciones de la diagonal W
 * @return Puntero al precondicionador diagonal
 */
double * 
PrecondDiag(const struct matrizmorse *A, const size_t *posdiag_A, const double *w, size_t *posdiag_W);

/**
 * @brief Función que calcula el precondicionador SSOR para matrices no simétricas
 * @param[in] A Matriz A del sistema
 * @param[in] posdiag_A Posiciones de la diagonal de la matriz A
 * @param[in] w Valor del precondicionador
 * @param[out] posdiag_W Posiciones de la diagonal W
 * @return Puntero al precondicionador diagonal
 */
double * 
PrecondSSOR_NS(const struct matrizmorse *A, const size_t *posdiag_A, const double *w, size_t *posdiag_W);

/**
 * @brief Función que calcula el precondicionador ILU
 * @param[in] A Matriz A del sistema
 * @param[in] posdiag_A Posiciones de la diagonal de la matriz A
 * @param[in] w Valor del precondicionador  (no se usa)
 * @param[out] posdiag_W Posiciones de la diagonal W
 * @return Puntero al precondicionador diagonal
 */
double * 
PrecondILU(const struct matrizmorse *A, const size_t *posdiag_A, const double *w, size_t *posdiag_W);

/**
 * @brief Función que calcula el precondicionador diagonal optimo
 * @param[in] A Matriz A del sistema
 * @param[in] posdiag_A Posiciones de la diagonal de la matriz A (no se usa)
 * @param[in] w Valor del precondicionador  (no se usa)
 * @param[out] posdiag_W Posiciones de la diagonal W
 * @return Puntero al precondicionador diagonal
 */
double * 
PrecondDiagOptimo(const struct matrizmorse *A, const size_t *posdiag_A, const double *w, size_t *posdiag_W);

/**
 * @brief Función que realiza el remonte para el precondicionador en el caso no simétrico
 * @param[in] r Vector a remontar
 * @param[in] W vector precondicionador
 * @param[in] posdiag_W vector de posiciones de la diagonal de W
 * @param[in] A Matrix del sistema
 * @param[out] y Vector con el resultado del remonte
 */
void 
RemonteLU(const double *r, const double *W, const size_t *posdiag_W, const struct matrizmorse *A, double *y);

/**
 * @brief Función que realiza el remonte para el precondicionador Diagonal
 * @param[in] r Vector a remontar
 * @param[in] W vector precondicionador
 * @param[in] posdiag_W vector de posiciones de la diagonal de W
 * @param[in] A Matrix del sistema
 * @param[out] y Vector con el resultado del remonte
 */
void 
RemonteDiag(const double *r, const double *W, const size_t *posdiag_W, const struct matrizmorse *A, double *y);

/**
 * @brief Función que realiza el remonte para el precondicionador Diagonal
 * @param[in] r Vector a remontar
 * @param[in] W vector precondicionador
 * @param[in] posdiag_W vector de posiciones de la diagonal de W
 * @param[in] A Matrix del sistema
 * @param[out] y Vector con el resultado del remonte
 */
void 
RemonteIdent(const double *r, const double *W, const size_t *posdiag_W, const struct matrizmorse *A, double *y);

/**
 * @brief Función que resuelve un sistema de ecuaciones con el metodo de gradientes conjugados
 *
 * @param matriz matriz del sistema (tipo morse)
 * @param[in] b vector de terminos independientes del sistema
 * @param[out] solucion vector para depositar la solucion del sistema
 * @param[in] u0 vector solucion inicial
 * @param[in] numiter numero maximo de iteraciones
 * @param[in] toler tolerancia entre soluciones sucesivas
 * @param[in] W vector precondicionador
 * @param[in] posdiag_W vector de posiciones de la diagonal de W
 * @param[in] remonte funcion a usar para el remonte
**/
bool GradienteConjugado(const struct matrizmorse *matriz, const double *b, double *solucion,
     const double *u0, const size_t numiter, const double toler, const double *W,
     const size_t *posdiag_W, const TFuncRemonte remonte);


/**
 * @brief Función que resuelve un sistema de ecuaciones con el metodo BiCGSTAB
 *
 * @param matriz matriz del sistema (tipo morse)
 * @param[in] b vector de terminos independientes del sistema
 * @param[out] solucion vector para depositar la solucion del sistema
 * @param[in] u0 vector solucion inicial
 * @param[in] numiter numero maximo de iteraciones
 * @param[in] toler tolerancia entre soluciones sucesivas
 * @param[in] W vector precondicionador
 * @param[in] posdiag_W vector de posiciones de la diagonal de W
 * @param[in] remonte funcion a usar para el remonte
**/
bool BICGSTAB(const struct matrizmorse *matriz, const double *b, double *solucion,
         const double *u0, const size_t numiter, const double toler, const double *W,
         const size_t *posdiag_W, const TFuncRemonte remonte);

/**
 * @brief Función que busca el vector columna i-esimo de la matriz A
 * @param[in] A Matriz en la que buscar
 * @param[in] col Número de columna
 * @param[out] vec Vector con los valores de la columna i-esima de A
 * @return Número de elementos del vector columna
 */
size_t BuscaColumna(const struct matrizmorse *A, const size_t col, double *vec);

/**
 * @brief Función que busca el elemento en la fila f, cuya columna sea c
 * @param[in] A Matriz en la que buscar
 * @param[in] f Número de fila en la que buscar
 * @param[in] c Número de columna
 * @return La posición en la matriz A si se encuentra, el valor de dim+1 en caso contrario
 */
size_t BuscaEnFila(const struct matrizmorse *A, const size_t f, const size_t c);

#endif //GRAD_H
