#ifndef INTERPOLACION_H
#define INTERPOLACION_H

#include <stddef.h>

/**
 * @file interpolacion.h
 * @brief Módulo para interpolar el campo de viento y los parámetros
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @brief Estructura con toda la información del modelo de predicción harmonie
 */
struct VariableHARMONIE
{
    /**
   * @brief Dimensiones de la coordenada X
   */
  size_t dimx;
  /**
   * @brief Dimensiones de la coordenada X
   */
  size_t dimy;
  /**
   * @brief Dimensiones de la coordenada Z
   */
  size_t dimz;

  /**
   * @brief Valor de la latitud en coordenadas UTM
   */
  double *latitud;
  /**
   * @brief Valor de la longitud en coordenadas UTM
   */
  double *longitud;
  /**
   * @brief Valor de la altura
   */
  double *altura;

  /**
   * @brief Componente U del viento
   */
  double *data_u;
  /**
   * @brief Componente V del viento
   */
  double *data_v;
  /**
   * @brief Componente W del viento
   */
  double *data_w;

  /**
   * @brief Menor de los incrementos de x (de longitud)
   */
  double deltaxmin;
  /**
   * @brief Menor de los incrementos de y (de latitud)
   */
  double deltaymin;

  /**
   * @brief Mayor de los incrementos de x (de longitud)
   */
  double deltaxmax;
  /**
   * @brief Mayor de los incrementos de y (de latitud)
   */
  double deltaymax;

  /**
   * @brief Maxima de las minimas coordenadas X
   */
  double xmaxmin;
  /**
   * @brief Maxima de las minimas coordenadas Y
   */
  double ymaxmin;

  /**
   * @brief Valor mínimo de la coordenada X
   */
  double xmin;
  /**
   * @brief Valor mínimo de la coordenada Y
   */
  double ymin;
  /**
   * @brief Valor máximo de la coordenada X
   */
  double xmax;
  /**
   * @brief Valor máximo de la coordenada Y
   */
  double ymax;
};


/**
 * @brief Función que lee un fichero con los parámetros de harmonie
 *        y los guarda en una estructura VariableHARMONIE
 * @param[in] nombreFicheroHARMONIE Nombre del fichero con los datos
 * @param[out] u Estructura de salida con los datos leídos
 */
void LeerVientoHARMONIE(const char *nombreFicheroHARMONIE,
            struct VariableHARMONIE *u);



/*
  Interpola7Vecinos


  Entradas:
  q : variables de WRF
  x, y, z : coordenadas UTM del punto donde se desea interpolar

  Salidas:
  vx, vy, vz : viento interpolado
*/
/**
 * @brief Interpola las tres componentes de viento a partir de las variables U, V, W de HARMONIE.
 * @param[in] q Variables del modelo Harmonie
 * @param[in] puntox,puntoy,puntoz Coordenada X Y Z (UTM) del punto que se quiere interpolar
 * @param[out] vx,vy,vz Componente X Y Z del viento interpolado
 */
void Interpola7VecinosHARM(const struct VariableHARMONIE *q,
               const double *puntox, const double *puntoy, const double *puntoz,
               double *vx, double *vy, double *vz);

/**
 * @brief Función que busca los valores de los incrementos de la malla harmonie
 * @param[inout] q Estructura de los valores de harmonie, se modifican las componentes deltaxmax, deltaxmin, deltaymax, deltaymin
 */
void BuscarIncrementosHARM(struct VariableHARMONIE *q);

/**
 * @brief Función que busca el bounding box de la malla harmonie
 * @param[inout] q Estructura de los valores de harmonie, se modifican las componentes xmin, xmax, ymin, ymax, xmaxmin, xmaxmin
 */
void BuscarBoundingBoxHARM(struct VariableHARMONIE *q);

/**
 * @brief Función que calcula la distancia entre el punto (utmx,utmy) y las correspondientes coordenadas utm la variable q en (i,j)
 * @param[in] q Variable con la estructura del harmonie
 * @param[in] utmx,utmy Punto UTM
 * @param[in] i,j Localización en la malla de Harmonie
 * @return
 */
double DistanciaR2ijHARM(const struct VariableHARMONIE *q, const double* utmx, const double* utmy,
        const int i, const int j);


/*
   Localiza la (i, j, k) mas cercana al punto dado en coordenadas UTM
Entradas:
 *q : puntero a la variable WRF en la que hacer la busqueda
 utmx, utmy, utmz: punto a localizar
t : instante de tiempo

Salidas:
 *i, *j, *k : posicion mas cercana de la rejilla

OJO: Version con ventana de busqueda; busca desde el i0,j0
hasta el i0+ki,j0+kj . Esto no siempre ha de funcionar.
Depende de lo deformada que este la rejilla

*/
/**
 * @brief Función que localiza la (i, j, k) mas cercana al punto dado en coordenadas UTM
 * @param[in] utmx,utmy,utmz Coordenadas UTM del punto a encontrar
 * @param[in] q Varialbe con la estructura del Harmonie
 * @param[out] qi,qj,qk Localización del punto del Harmonie más cercano a las cordenadas UTM
 */
void LocalizarEnRejillaHARM(const double* utmx, const double* utmy, const double* utmz,
        const struct VariableHARMONIE *q,
        int *qi, int *qj, int *qk);

// Dados los indices i, j, k devuelve las coordenas UTM correspondientes
// de la variable *q
// i para la Y (latitud)
// j para la X (longitud)
// k para la altura
/**
 * @brief Funciónq que devuelve las coordenadas UTM de la localización en el Harmonie
 * @param[in] q
 * @param[in] i,j,k Localización en la malla Harmonie
 * @param[out] utmx,utmy,utmz Coordenadas UTM del punto a localizar
 */
void Utm3DHARM(const struct VariableHARMONIE *q, size_t i, size_t j, size_t k,
        double *utmx, double *utmy, double *utmz);

/**
 * @brief Función que interpola el viento con 7 vecinos de la malla del Harmonie
 * @param[in] q Variable con toda la estructura del Harmonie
 * @param[in] x,y,z Coordenadas UTM del punto donde se desea interpolar
 * @param[out] wind Valor interpolado
 */
void InterpolaComponente7VecinosHARM(const struct VariableHARMONIE *q, const double* x, const double* y,
        const double* z, double wind[3]);

#endif //INTERPOLACION_H
