#include "lectura.h"

#include <errno.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

/**
 * @file lectura.c
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

void parseLineAsListOfDoubles(const size_t longitud, char *buffer, const char *nombreFichero,
                              const size_t line_number, double *array)
{
    size_t i=0;
    char *buf_dummy = buffer;
    while(*buf_dummy) {
        char *p;
        array[i] = strtod(buf_dummy, &p);
        i++;
        if (buf_dummy == p) {
            fprintf(stderr, "%s: not a double number\n", buf_dummy);
            fprintf(stderr, "%s:%zu\n",nombreFichero, line_number);
            exit(EXIT_FAILURE);
        } else if (('\0' != *p) && (i > longitud-1)) {
            break;
        } else {
            buf_dummy = p + 1;
        }

    }
    if (i != longitud){
        fprintf(stderr, "Error leyendo los valores de la linea. No hay %zu datos\n", longitud);
        fprintf(stderr, "%s:%zu\n",nombreFichero, line_number);
        exit(EXIT_FAILURE);
    }
}

void parseLineAsListOfUInts(const size_t longitud, char *buffer, const char *nombreFichero,
                            const size_t line_number, size_t *array)
{
    size_t i=0;
    char *buf_dummy = buffer;

    while(*buf_dummy) {
        char *p;
        array[i] = strtoul(buf_dummy, &p, 10);
        i++;
        if (buf_dummy == p) {
            fprintf(stderr, "%s: not a decimal number\n", buf_dummy);
            fprintf(stderr, "%s:%zu\n", nombreFichero, line_number);
            exit(EXIT_FAILURE);
        } else if (('\0' != *p) && (i > longitud-1)) {
            break;
        } else {
            buf_dummy = p + 1;
        }
    }

    if (i != longitud){
        fprintf(stderr, "Error leyendo los valores de la linea, no hay %zu datos\n", longitud);
        fprintf(stderr, "%s:%zu\n", nombreFichero, line_number);
        exit(EXIT_FAILURE);
    }
}


int LeeLinea(FILE *f, char *buffer, const size_t buffersize, size_t *line_number)
{

    // Se salta las lineas vacias o con comentarios
    // Los comentarios tienen el caracter # en la columna 0
    do {
        if (fgets(buffer, buffersize, f) != NULL) {
            ++(*line_number);
            char *p = strchr(buffer, '\n');
            if (p) {
                *p = '\0';
            } else {
                fprintf(stderr, "Line %zu longer than buffer\n", *line_number);
                fprintf(stderr, "line content: %s\n", buffer);
                return 1 ;
            }
        } else {
            *buffer = '\0';
            return 2 ;
        }
    } while (((buffer[0]=='#') || (buffer[0]=='\0')));
    return 0 ;
}// LeeLinea


void LeerParametros(const char *datos, struct T_DatosAtmos *datmos)
{

    struct param_ga paramet [] = {
    {"CTE_VON_KARMAN",                DOUBLE,    false, ""},
    {"PASQUILL",                      INT,      false, ""},
    {"V_GEOSTROFICO_X",               DOUBLE,    false, ""},
    {"V_GEOSTROFICO_Y",               DOUBLE,    false, ""},
    {"V_GEOSTROFICO_Z",               DOUBLE,    false, ""},
    {"ALPHA",                         DOUBLE,    false, ""},
    {"EPSILON",                       DOUBLE,    false, ""},
    {"GAMMA1",                        DOUBLE,    false, ""},
    {"GAMMA2",                        DOUBLE,    false, ""},
    {"LATITUD",                       DOUBLE,    false, ""},
    {"FICH_VIENTO",                   STRING,   false, ""},
    {"FICH_MALLA",                    STRING,   false, ""},
};

    const size_t NUM_PARAM_PARAMET = 12;

    size_t i, idlen, line_number = 0;
    int par_int=-1;
    FILE *fp;
    const size_t tambuf = 256;
    char buf[tambuf], *pos;
    double par_flt=-1.0;

    // Inicializa algunos valores de datmos
    datmos->fich_malla = NULL;
    datmos->tipo_fich_viento = 0; // Por defecto se leen ficheros de las
    // estaciones  con los datos de viento en componentes polares

    // Abre el fichero para leer
    if(fp=fopen(datos, "r"),!fp){
        fprintf(stderr, "LeerParametros:Problema al abrir el fichero %s: %s\n",
                datos, strerror(errno));
        exit(EXIT_FAILURE);
    }

    while (LeeLinea(fp, buf, tambuf, &line_number) == 0) {
        for (i = 0; i < NUM_PARAM_PARAMET; i++){
            idlen = strlen(paramet[i].idstring);
            if (strncmp(buf, paramet[i].idstring, idlen) == 0) {
                if (paramet[i].set){
                    fprintf(stderr, "%s definido repetidamente\n",
                            paramet[i].idstring);
                    fprintf(stderr, "%s:%zu\n", datos, line_number);
                    exit(EXIT_FAILURE);
                }
                break;
            }
        }

        if (i == NUM_PARAM_PARAMET) {
            fprintf(stderr, "Atencion: Parametro [%s] no reconocido\n",
                    buf);
            fprintf(stderr, "%s:%zu\n", datos, line_number);
            continue;
        }

        paramet[i].set = true;
        pos = buf + idlen;
        while ((*pos != '=') && (*pos != '\0')) pos++;
        if (*pos == '=')
            pos++;
        else {
            fprintf(stderr, "No se encontro '=' para %s\n", paramet[i].idstring);
            exit(EXIT_FAILURE);
        }
        while (*pos <=32) pos++ ;
        strcpy(paramet[i].data, pos);
    }// fin del while


    for (i = 0; i < NUM_PARAM_PARAMET; i++){
        if (paramet[i].set) {
            switch(paramet[i].type)
            {
            case INT:
                par_int = atoi(paramet[i].data);
                break;
            case STRING:
                break;
            case DOUBLE:
                par_flt = atof(paramet[i].data);
                break;
            }// switch

            if ( strcmp(paramet[i].idstring, "LATITUD") == 0 ) {
                datmos->dlat = par_flt * PI / 180.0;
            } else if ( strcmp(paramet[i].idstring,  "CTE_VON_KARMAN") == 0 ) {
                datmos->dk = par_flt;
            } else if ( strcmp(paramet[i].idstring,  "PASQUILL") == 0 ) {
                datmos->pasquill = par_int;
            } else if ( strcmp(paramet[i].idstring,  "V_GEOSTROFICO_X") == 0 ) {
                datmos->vg.x = par_flt;
            } else if ( strcmp(paramet[i].idstring,  "V_GEOSTROFICO_Y") == 0 ) {
                datmos->vg.y = par_flt;
            } else if ( strcmp(paramet[i].idstring,  "V_GEOSTROFICO_Z") == 0 ) {
                datmos->vg.z = par_flt;
            } else if ( strcmp(paramet[i].idstring,  "ALPHA") == 0 ) {
                datmos->alpha = par_flt;
            } else if ( strcmp(paramet[i].idstring,  "EPSILON") == 0 ) {
                datmos->epsilon = par_flt;
            } else if ( strcmp(paramet[i].idstring,  "GAMMA1") == 0 ) {
                datmos->gamma = par_flt;
            } else if ( strcmp(paramet[i].idstring,  "GAMMA2") == 0 ) {
                datmos->gamma2 = par_flt;
            } else if ( strcmp(paramet[i].idstring,  "FICH_VIENTO") == 0 ) {
                datmos->fich_viento = (char *)malloc(strlen(paramet[i].data) + 1);
                if (datmos->fich_viento != NULL) {
                    strcpy(datmos->fich_viento, paramet[i].data);
                } else {
                    fprintf(stderr, "LeerParametros: %s: %s\n",
                            paramet[i].idstring, strerror(errno));
                    exit(EXIT_FAILURE);
                }
            } else if ( strcmp(paramet[i].idstring,  "FICH_MALLA") == 0 ) {
                datmos->fich_malla = (char *)malloc(strlen(paramet[i].data) + 1);
                if (datmos->fich_malla != NULL) {
                    strcpy(datmos->fich_malla, paramet[i].data);
                } else {
                    fprintf(stderr, "LeerParametros: %s: %s\n",
                            paramet[i].idstring, strerror(errno));
                    exit(EXIT_FAILURE);
                }
            } else {
                fprintf(stderr,
                        "LeerParametros: error 2: parametro [%s]\n",
                        paramet[i].idstring);
                exit(EXIT_FAILURE);
            }
        }// if
    } // for

    fclose(fp);

    datmos->omega = 7.292115055e-5;
    datmos->f = 2.0 * datmos->omega * sin(datmos->dlat);
    datmos->altest = 10.0; // Se suponen todas las estac. a 10m de altura
    datmos->t = 0; // De momento no se usa
    datmos->metodo = 0;
    datmos->precond = 2;

    // Comprueba que se hayan definido todos los ficheros de la malla
    if (datmos->fich_malla == NULL){
        fprintf(stderr, "LeerParametros: falta el fichero de coordenadas de la malla\n");
        exit(EXIT_FAILURE);
    }

    // Comprueba qye se haya definido el fichero de la topografia
    if (!datmos->tieneRugosidad && datmos->fich_topo == NULL){
        fprintf(stderr, "LeerParametros: falta el fichero de la topografia\n");
        exit(EXIT_FAILURE);
    }


}// LeerParametros


/**
 * Lee los ficheros de coordenadas y elementos de los tetraedros
 * de una malla; asigna memoria para el viento inicial (vini) y
 * y el final (wind).
 */

int LeerMalla(const char *f_malla, struct T_Malla *Malla)
{
    FILE *fp;
    char buffer[256];
    size_t line_number = 0;

    if(fp=fopen(f_malla, "r"),!fp){
        fprintf(stderr, "LeerMalla:Problema con la apertura del fichero de coordenadas %s: %s\n",
                f_malla, strerror(errno));
        exit(EXIT_FAILURE);
    }
    // Reserva memoria para las coordenadas de los nodos y las
    // condiciones de contorno
    if (LeeLinea(fp, buffer, sizeof(buffer), &line_number) == 0) {
        char *buf_dummy = buffer;
        char *p;
        Malla->Num_nodos = strtoul(buf_dummy, &p, 10);
        if (buf_dummy == p) {
            fprintf(stderr, "%s: not a decimal number\n", buf_dummy);
            fprintf(stderr, "%s:%zu\n", f_malla, line_number);
            exit(EXIT_FAILURE);
        } else {
            buf_dummy = p + 1;
        }
        Malla->Num_tetra = strtoul(buf_dummy, &p, 10);
        if (buf_dummy == p) {
            fprintf(stderr, "%s: not a decimal number\n", buf_dummy);
            fprintf(stderr, "%s:%zu\n", f_malla, line_number);
            exit(EXIT_FAILURE);
        } else if ('\0' != *p) {
            fprintf(stderr, "%s: extra characters at end of input: %s\n", buffer, p);
            fprintf(stderr, "%s:%zu\n", f_malla, line_number);
            exit(EXIT_FAILURE);
        } else {
            buf_dummy = p + 1;
        }
    }

    if ((Malla->nodo = (double **)malloc(sizeof(double*) * Malla->Num_nodos))
            == NULL){
        fprintf(stderr,"LeerMalla: no puedo reservar memoria para malla->nodo\n");
        exit(EXIT_FAILURE);
    }
    for (size_t i=0; i< Malla->Num_nodos; i++)
        if ((Malla->nodo[i] = (double *)malloc(sizeof(double) * 3)) == NULL){
            fprintf(stderr, "LeerMalla: no puedo reservar memoria para malla->nodo[%zu]\n", i);
            exit(EXIT_FAILURE);
        }
    if ((Malla->condicion = (short int *)malloc(sizeof(short int) * Malla->Num_nodos))
            == NULL){
        fprintf(stderr, "LeerMalla: no puedo reservar memoria para malla->condicion\n");
        exit(EXIT_FAILURE);
    }

    if ((Malla->tetra = (size_t **)malloc(sizeof(size_t*) * Malla->Num_tetra))
            == NULL){
        fprintf(stderr, "LeerMalla: no puedo reservar memoria para malla->Num_tetra\n");
        exit(EXIT_FAILURE);
    }
    for (size_t i=0; i< Malla->Num_tetra; i++)
        if ((Malla->tetra[i] = (size_t *)malloc(sizeof(size_t) * 4)) == NULL){
            fprintf(stderr, "LeerMalla: no puedo reservar memoria para malla->tetra[%zu]", i);
            exit(EXIT_FAILURE);
        }
    // Reserva memoria para el viento inicial en cada elemento
    if ((Malla->vini = (double **)malloc(sizeof(double*) * Malla->Num_tetra))
            == NULL){
        fprintf(stderr, "LeerMalla: no puedo reservar memoria para malla->vini\n");
        exit(EXIT_FAILURE);
    }
    for (size_t i=0; i< Malla->Num_tetra; i++)
        if ((Malla->vini[i] = (double *)malloc(sizeof(double) * 3)) == NULL){
            fprintf(stderr, "LeerMalla: no puedo reservar memoria para malla->vini[%zu]\n", i);
            exit(EXIT_FAILURE);
        }

    // Reserva memoria para el viento calculado en cada elemento
    if ((Malla->wind = (double **)malloc(sizeof(double*) * Malla->Num_tetra))
            == NULL){
        fprintf(stderr, "LeerMalla: no puedo reservar memoria para malla->wind\n");
        exit(EXIT_FAILURE);
    }
    for (size_t i=0; i< Malla->Num_tetra; i++)
        if ((Malla->wind[i] = (double *)malloc(sizeof(double) * 3)) == NULL){
            fprintf(stderr, "LeerMalla: no puedo reservar memoria para malla->wind[%zu]\n", i);
            exit(EXIT_FAILURE);
        }

    // Reserva memoria para las coordenadas de los centros de los tetraedros
    if ((Malla->cen_tetra=(double **)malloc(sizeof(double*) * Malla->Num_tetra))
            == NULL){
        fprintf(stderr, "LeerMalla: no puedo reservar memoria para malla->cen_tetra\n");
        exit(EXIT_FAILURE);
    }
    for (size_t i=0; i< Malla->Num_tetra; i++){
        if ((Malla->cen_tetra[i] = (double *)malloc(sizeof(double) * 3)) == NULL){
            fprintf(stderr, "LeerMalla: no puedo reservar memoria para malla->cen_tetra[%zu]", i);
            exit(EXIT_FAILURE);
        }
    }

    Malla->z_terr = (double *)malloc(Malla->Num_tetra* sizeof(double));
    if (Malla->z_terr == NULL){
        fprintf(stderr, "Error alocatando Malla->z_terr\n");
        exit(EXIT_FAILURE);
    }

    // Lectura de las coordenadas de los nodos y las
    // condiciones de contorno
    {
        size_t counter_nodo = 0;
        while (counter_nodo < Malla->Num_nodos) {
            if (LeeLinea(fp, buffer, sizeof(buffer), &line_number) == 0) {

                parseLineAsListOfDoubles(3, buffer, f_malla, line_number, Malla->nodo[counter_nodo]);
                Malla->condicion[counter_nodo] = 0;

                counter_nodo++;
            } else {
                fprintf(stderr,"LeerMalla: Problema al leer el fichero de nodos\n");
                fprintf(stderr, "%s:%zu\n", f_malla, line_number);
                exit(EXIT_FAILURE);
            }
        }

    }

    // Lectura de los nodos que componen cada tetraedro
    {
        size_t counter_tetra = 0;
        bool oneIndexed = true;
        while (counter_tetra < Malla->Num_tetra){
            if (LeeLinea(fp, buffer, sizeof(buffer), &line_number) == 0) {
                char *buf_dummy = buffer;
                size_t i = 0;
                while(*buf_dummy) {
                    char *p;
                    Malla->tetra[counter_tetra][i] = strtoul(buf_dummy, &p, 10);
                    if (Malla->tetra[counter_tetra][i] == 0) {
                        oneIndexed = false;
                    }
                    i++;
                    if (buf_dummy == p) {
                        fprintf(stderr, "%s: not an integer number\n", buf_dummy);
                        fprintf(stderr, "%s:%zu\n", f_malla, line_number);
                        exit(EXIT_FAILURE);
                    } else if (('\0' != *p) && (i > 3)) {
                        break;
                    } else {
                        buf_dummy = p + 1;
                    }

                }

                if (i != 4){
                    fprintf(stderr, "Error leyendo las conectividades de la malla, no hay 4\n");
                    fprintf(stderr, "%s:%zu\n", f_malla, line_number);
                    exit(EXIT_FAILURE);
                }

                counter_tetra++;

            } else {
                fprintf(stderr,"LeerMalla: Problema al leer el fichero de elementos\n");
                fprintf(stderr, "%s:%zu\n", f_malla, line_number);
                exit(EXIT_FAILURE);
            }
        }


        fclose(fp);

        if (oneIndexed) {
            for (size_t i = 0; i<Malla->Num_tetra; i++) {
                for (size_t j=0; j<4; j++) {
                    Malla->tetra[i][j] = Malla->tetra[i][j]-1;
                }
            }
        }
    }


    // Multiplicamos y dividimos por mil para redondear. La malla de Manel
    // tenia valores por debajo del x_min (seguramente tambien y_min, etc.)
    for (size_t i = 0; i < Malla->Num_nodos; i++) {
        Malla->nodo[i][0] = round(Malla->nodo[i][0]*1000.00)/1000.00;
        Malla->nodo[i][1] = round(Malla->nodo[i][1]*1000.00)/1000.00;
        Malla->nodo[i][2] = round(Malla->nodo[i][2]*1000.00)/1000.00;
    }


    return(Malla->Num_nodos);
}// LeerMalla
