#ifndef LECTURA_H
#define LECTURA_H

//#include "topografia.h"
//#include "localizador.h"
#include "wind3D.h"

#include <stdbool.h>
#include <stdio.h>


/**
 * @file lectura.h
 * @brief Métodos relacionados con la lectura de la malla
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
  * @brief Enumerador para el tipo de valor en el fichero de configuración
  */
enum read_type {INT, STRING, DOUBLE};


/**
 * @brief Estructura para leer el archivo de configuración
 */
struct param_ga {
    /**
   * @brief Identificador que se usa en el archivo de configuración
   */
  char *idstring;
  /**
   * @brief Tipo de valor del parámetro (real, entero, cadena, ...)
   */
  enum read_type type;
  /**
   * @brief Verdadero si el valor ya se ha asignado, falso en caso contrario
   */
  bool set;
  /**
   * @brief Cadena temporal donde guardar el dato leido para su futura traslación
   */
  char data[128];
};

/**
 * @brief Función que dada una cadena de caracteres, extrae n reales y los asigna al vector array
 * @param[in] longitud Número de valores a leer en la línea
 * @param[inout] buffer Cadena que contiene la línea
 * @param[in] nombreFichero Nombre del fichero, usado para el mensaje de error
 * @param[in] line_number Número de línea, usado para el mensaje de error
 * @param[out] array Vector en el que se asigna el valor del contenido de la línea
 */

void parseLineAsListOfDoubles(const size_t longitud, char *buffer, const char *nombreFichero,
                              const size_t line_number, double *array);

/**
 * @brief Función que dada una cadena de caracteres, extrae n enteros y los asigna al vector array
 * @param[in] longitud Número de valores a leer en la línea
 * @param[inout] buffer Cadena que contiene la línea
 * @param[in] nombreFichero Nombre del fichero, usado para el mensaje de error
 * @param[in] line_number Número de línea, usado para el mensaje de error
 * @param[out] array Vector en el que se asigna el valor del contenido de la línea
 */
void parseLineAsListOfUInts(const size_t longitud, char *buffer, const char *nombreFichero,
                            const size_t line_number, size_t *array);

/**
 * @brief Función que sirve para leer una línea de un fichero
 * @param[in] f Fichero a leer
 * @param[out] buffer Línea resultante
 * @param[in] buffersize Tamaño de la línea
 * @param[inout] line_number Variable que guarda el número de la línea del fichero
 * @return 0 en caso de lectura válida, 1 en caso contrario
 */

int
LeeLinea(FILE *f, char *buffer, const size_t buffersize, size_t *line_number);


/**
 * @brief Lee los parametros del problema: datos geograficos, atmosfericos,
 * y parametros del modelo.
 *
 * @param[in] datos nombre del fichero de parametros
 * @param[out] datmos puntero a estructura de datos donde guardar parametros
 */

void
LeerParametros(const char *datos, struct T_DatosAtmos *datmos);

/**
 * @brief Lee los ficheros de coordenadas y elementos de los tetraedros
 *        de una malla.
 *
 * @param[in] f_malla Nombre del fichero de malla
 * @param[out] Malla Estructura de malla a rellenar con los datos leídos
 * @return Número de nodos en la malla leída
 */

int
LeerMalla(const char *f_malla, struct T_Malla *Malla);

/**
 * @brief Función que libera la memoria usada por la estructura de la variable Malla
 * @param[inout] Malla Variable a liberar memoria
 */

void
LiberarMalla(struct T_Malla *Malla);

#endif //LECTURA_H
