#ifndef LOCALIZADOR_H
#define LOCALIZADOR_H

//#include "lectura.h"

#include <stdbool.h>
#include <stddef.h>

/**
 * @file localizador.h
 * @brief Módulo para localizar rápidamente puntos en mallas de triángulos o tetraedros
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @brief Estructura de árbol base para facilitar las busquedas en la malla, se usará como octree y como quadtree.
 */
struct tree{
    /**
     * @brief branch Lista de elementos tree para recursividad
     */
    struct tree* branch;
    /**
     * @brief nOfElements nOfElements Número de elementos que contiene la hoja
     */
    size_t nOfElements;
    /**
     * @brief maxNOfElements
     */
    size_t maxNOfElements;
    /**
     * @brief level
     */
    size_t level;
    /**
     * @brief maxLevel
     */
    size_t maxLevel;
    /**
     * @brief isRoot isRoot Boleano que indica si la hoja contiene hijos
     */
    bool isRoot;
    /**
     * @brief elements Lista de los elementos contenidos en la hoja del árbol
     */
    size_t* elements;
    /**
     * @brief x_min
     */
    double x_min;
    /**
     * @brief x_max
     */
    double x_max;
    /**
     * @brief y_min
     */
    double y_min;
    /**
     * @brief y_max
     */
    double y_max;
    /**
     * @brief z_min
     */
    double z_min;
    /**
     * @brief z_max
     */
    double z_max;
};

/**
 * @brief Estructura que guarda los datos del localizador en 2D
 */

struct sLocation2D {
    /**
     * @brief Número de intervalos en la coordenada X
     */
    size_t nx;
    /**
     * @brief Número de intervalos en la coordenada Y
     */
    size_t ny;
    /**
     * @brief Valor máximo en la coordenada X
     */
    double xMax;
    /**
     * @brief Valor máximo en la coordenada Y
     */
    double yMax;
    /**
     * @brief Valor mínimo en la coordenada X
     */
    double xMin;
    /**
     * @brief Valor mínimo en la coordenada Y
     */
    double yMin;
    /**
     * @brief Número de tetraedros que caen en una posición de la cuadrícula
     */
    size_t **size;
    /**
     * @brief Id del tetraedro que cae en una cuadrícula
     */
    size_t ***location;
};

/**
 * @brief Estructura que guarda los datos del localizador en 2D
 */

struct sLocation3D {
    /**
     * @brief Número de intervalos en la coordenada X
     */
    size_t nx;
    /**
     * @brief Número de intervalos en la coordenada Y
     */
    size_t ny;
    /**
     * @brief Número de intervalos en la coordenada Z
     */
    size_t nz;
    /**
     * @brief Valor máximo en la coordenada X
     */
    double xMax;
    /**
     * @brief Valor máximo en la coordenada Y
     */
    double yMax;
    /**
     * @brief Valor máximo en la coordenada Z
     */
    double zMax;
    /**
     * @brief Valor mínimo en la coordenada X
     */
    double xMin;
    /**
     * @brief Valor mínimo en la coordenada Y
     */
    double yMin;
    /**
     * @brief Valor mínimo en la coordenada Z
     */
    double zMin;
    /**
     * @brief Número de tetraedros que caen en una posición de la cuadrícula
     */
    size_t ***size;
    /**
     * @brief Id del tetraedro que cae en una cuadrícula
     */
    size_t ****location;
};


/**
 * @brief Función que comprueba si un triangulo se intersecta con la bounding box del quadtree
 * @param[in] situation Variable con la estructura del quadtree
 * @param[in] coor Vector de coordenadas
 * @param[in] conec Vector de conectividades del triangulo
 * @return Devuelve verdadero si se intersecta, falso en caso contrario
 */
bool tringleIsInsideQT(struct tree *situation, const double **coor, const size_t *conec);

/**
 * @brief Función que genera todo el quadtree, como localizador para una malla de triángulos
 * @param[in] x_min,y_min,z_min Valores mínimos de X Y Z
 * @param[in] x_max,y_max,z_max Valores máximos de X Y z
 * @param[in] coor Vector de coordenadas de la malla
 * @param[in] conec Vector de conectividades de la malla
 * @param[in] numCaras Número de triangulos de la malla
 * @param[out] situation Variable quadtree procesada
 */
void getLocation_quadTree(const double x_min, const double y_min, const double z_min,
                          const double x_max, const double y_max, const double z_max,
                          const double **coor, const size_t **conec, const size_t numCaras,
                          struct tree *situation);

/**
 * @brief Función que genera todo el octtree, como localizador para una malla de tetraedros
 * @param[in] x_min,y_min,z_min Valores mínimos de X Y Z
 * @param[in] x_max,y_max,z_max Valores máximos de X Y z
 * @param[in] coor Vector de coordenadas de la malla
 * @param[in] conec Vector de conectividades de la malla
 * @param[in] numElems Número de tetraedros de la malla
 * @param[out] situation Variable octtree procesada
 */
void getLocation_OctTree(const double x_min, const double y_min, const double z_min,
                         const double x_max, const double y_max, const double z_max,
                         const double **coor, const size_t **conec,
                         const size_t numElems, struct tree *situation);

/**
 * @brief Función que refina un quadtree para conseguir un localizador
 * @param[inout] situation Variable con el quadtree para localizar
 * @param[in] coor Coordenadas de la malla en la que localizar
 * @param[in] conec Conectividades de la malla en la que localizar
 */
void refine_quadTree(struct tree *situation, const double **coor, const size_t **conec);

/**
 * @brief Función que genera un localizador para mallas de tetraedros
 * @param[in] nx,ny,nz Número de intervalos en la coordenada X
 * @param[in] x_min,y_min,z_min Valor mínimo de la coordenada X Y Z
 * @param[in] x_max,y_max,z_max Valor máximo de la coordenada X Y Z
 * @param[in] coor Coordenadas de la malla
 * @param[in] conec Conectividades de la malla
 * @param[in] numTetra Número de tetraedros
 * @param[out] situation Estructura de salida donde se recoge toda la información del localizador
 */
void getLocation_tetrahedron(const size_t nx, const size_t ny, const size_t nz,
                             const double x_min, const double y_min, const double z_min,
                             const double x_max, const double y_max, const double z_max,
                             double **coor, size_t **conec, const size_t numTetra,
                             struct sLocation3D *situation);


/**
 * @brief Función que genera un localizador para superfícies de triángulos
 * @param[in] nx,ny Número de intervalos en la coordenada X
 * @param[in] x_min,y_min Valor mínimo de la coordenada X Y
 * @param[in] x_max,y_max Valor máximo de la coordenada X Y
 * @param[in] coor Coordenadas de la superfície
 * @param[in] conec Conectividades de la superficie
 * @param[in] numCaras Número de caras en la superficie
 * @param[out] situation Estructura de salida donde se recoge toda la información del localizador
 */

void getLocation_triangle(const size_t nx, const size_t ny, const double x_min, const double y_min,
                          const double x_max, const double y_max, double **coor, size_t **conec,
                          const size_t numCaras, struct sLocation2D *situation);

/**
 * @brief Función que dado un punto (x,y,z) busca en el localizador el tetraedro al que pertenece
 * @param[in] x,y,z Coordenada X Y Z del punto
 * @param[in] coord Coordenadas de la superfície
 * @param[in] conec Conectividades de la superficie
 * @param[in] numberElems Número de elementos
 * @param[in] situation Estructura que guarda la información del localizador
 * @return El número del triángulo en el que se encuentra el punto,
 *         o el número de caras de la malla +1 en caso de no encontrar triángulo
 */
size_t getTetrahedra(const double x, const double y, const double z,
                     const double **coord, const size_t **conec, const size_t numberElems,
                     const struct sLocation3D *situation);

/**
 * @brief Función que dado un punto (x,y,z) busca en el localizador el triangulo al que pertenece
 * @param[in] x,y Coordenada X Y del punto
 * @param[in] coord Vector con las coordenadas de los elementos
 * @param[in] conec Vector de conectividades de los elementos
 * @param[in] numElems Número de elementos de la malla (para poder devolver numElems+1)
 * @param[in] situation Estructura que guarda la información del localizador
 * @return El número del triángulo en el que se encuentra el punto,
 *         o el numElemes+1 en caso de no encontrar triángulo
 */
size_t getTriangle(const double x, const double y, const double **coord, const size_t **conec, const size_t numElems,
                   const struct sLocation2D *situation);

/**
 * @brief Función que dado un punto (x,y,z) busca en el localizador quadtree el triangulo al que pertenece
 * @param[in] x,y Coordenada X Y del punto
 * @param[in] coord Vector con las coordenadas de los elementos
 * @param[in] conec Vector de conectividades de los elementos
 * @param[in] numElems Número de elementos de la malla (para poder devolver numElems+1)
 * @param[in] situation Estructura que guarda la información del localizador
 * @return El número del triángulo en el que se encuentra el punto,
 *         o el numElemes+1 en caso de no encontrar triángulo
 */
size_t getTriangle_QT(const double x, const double y, const double **coord, const size_t **conec, const size_t numElems,
                      const struct tree *situation);

/**
 * @brief Función que busca si un punto está en un tetraedro (3D)
 * @param[in] p1,p2,p3,p4 Los cuatro vértices del tetraedro
 * @param[in] punto Punto que verificar
 * @return Verdadero si el punto está dentro del tetraedro,
 *         falso en caso contrario
 */

bool point_inside_tetrahedra(const double *p1, const double *p2,
                             const double *p3, const double *p4,
                             const double *punto);

/**
 * @brief Función que busca si un punto está en un triángulo (2D)
 * @param[in] p1,p2,p3 Los tres vértices del triángulo
 * @param[in] punto Punto que verificar
 * @return Verdadero si el punto está dentro del triángulo,
 *         falso en caso contrario
 */

bool point_inside_triangle(const double *p1, const double *p2, const double *p3, const double *punto) ;

/**
 * @brief Función que libera la memoria de la estructura quadtree
 * @param[inout] situation Variable a liberar
 */

void freeQT(struct tree *situation);

/**
 * @brief Función que calcula el bounding box de los hijos de un quadtree
 * @param[inout] situation Variable con la estructura quadtree
 */
void set_bb_branches(struct tree *situation);

/**
 * @brief Función que reparte los elementos del padre al hijo i-esimo en un quadtree
 * @param[in] conec Vector con las conectividades de los elementos
 * @param[in] coor Vector con las coordenadas de los elementos
 * @param[inout] situation Quadtree en el que se reparten los elementos
 * @param[in] i Número de hijo del quadtree
 */
void passElements(const size_t **conec, const double **coor, struct tree *situation, const size_t i);

/**
 * @brief Función que dado un punto XY y un quadtree busca todos los elementos de la hoja que contiene el punto XY
 * @param[in] x,y Coordenadas X Y del punto
 * @param[in] situation Variable del quadtree
 * @param[out] elements Vector con los elementos que contiene la hoja
 * @param[out] nOfElements Número de elementos
 */
void getElementsQT(const double x, const double y,
                   const struct tree *situation,
                   size_t **elements, size_t *nOfElements);

/**
 * @brief Función que reparte los elementos del padre al hijo i-esimo en un Octree
 * @param[in] conec Vector con las conectividades de los elementos
 * @param[in] coor Vector con las coordenadas de los elementos
 * @param[inout] situation Quadtree en el que se reparten los elementos
 * @param[in] i Número de hijo del quadtree
 */
void passElementsOT(const size_t **conec, const double **coor, struct tree *situation, size_t i);

/**
 * @brief Función que refina un octree para conseguir un localizador de tetraedros
 * @param[inout] situation Variable con el octree para localizar
 * @param[in] coor Coordenadas de la malla en la que localizar
 * @param[in] conec Conectividades de la malla en la que localizar
 */
void refine_OctTree(struct tree *situation, const double **coor, const size_t **conec);

/**
 * @brief Función que calcula el bounding box de los hijos de un octree
 * @param[inout] situation Variable con la estructura octtree
 */
void set_bb_branches_OT(struct tree *situation);

/**
 * @brief Función que pasa los valores de los parámetros de un nodo de padre a hijo
 * @param situation nodo padre
 * @param i número de hijo
 */
void inheritParametres(struct tree *situation, const size_t i);

#endif //LOCALIZADOR_H
