#include "wind3D.h"

#include "avs.h"
#include "capas.h"
#include "caras_ext.h"
#include "localizador.h"
#include "puntosSalida.h"
#include "viento.h"

#include <ctype.h>
#include <getopt.h>
#include <stdlib.h>
#include <string.h>

/**
 * @file main.c
 * @brief Fichero de ejecución de la librería wind3D
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


/**
 * @brief Muestra un mensaje explicando como usar el programa
 */
void uso()
{
    fprintf(stdout, "wind3D\n");
    fprintf(stdout, "Parametros obligatorios:\n");
    fprintf(stdout, "\t-c fichero de configuracion\n");
    fprintf(stdout, "\t-f Formato de entrada (forecast | stations)\n");
    fprintf(stdout, "\t-r Valor de rugosidad constante (no compatible con -R)\n");
    fprintf(stdout, "\t-R Nombre del fichero de rugosidad (no compatible con -r)\n");
    fprintf(stdout, "Parametros opcionales:\n");
    fprintf(stdout, "\t-s Nombre del fichero con los puntos de salida del modelo\n");
    fprintf(stdout, "\t-i Escribir fichero inp del viento interpolado\n");
    fprintf(stdout, "\t-n No escribir ficheros inp del viento resultante\n");
    fprintf(stdout, "\t-l Calcular el viento en las diferentes capas (10, 20, 50, 100, 150, 200, 300, 400, 500)\n");
    fprintf(stdout, "Uso:\n");
    fprintf(stdout, "   wind3D -c config.dat -f forecast|stations|interpolation -r num |-R fichero [-s fichero] [-l]\n");
} // uso

/**
 * @brief Función que procesa las opciones que se pasan en la línea de comando
 * @param[in] argc Número de argumentos en la línea de comandos
 * @param[in] argv Contenido de los argumetnos de la línea de comandos
 * @param[out] fich_param Fichero de configuración
 * @param[out] file_format Tipo de problema: stations forecast interpolation
 * @param[out] datosAtmos Variable con los datos atmosféricos
 * @param[out] malla Variable con la malla
 * @param[out] salidaInter Boleano que indica si se quiere escribir el viento interpolado
 * @param[out] escribir Boleano que indica si se quiere escribir los inp del campo de viento resultante
 */
void procesaArgumentos(int argc, char **argv, char *fich_param, char
                       *file_format, struct T_DatosAtmos *datosAtmos,
                       struct T_Malla *malla, bool *salidaInter, bool *escribir)
{
    int opcion;
    int cflag=0, fflag=0, rflag =0, Rflag=0;

    if (argc < 4) {
        fprintf(stderr, "Error: falta algun parametro obligatorio\n");
        uso();
        exit(EXIT_FAILURE);
    }


    while ((opcion = getopt(argc, argv, "c:f:r:R:s:lin")) != -1) {
        switch (opcion) {
        case 'c':
            strcpy(fich_param, optarg);
            cflag++ ;
            break;
        case 'f':
            strcpy(file_format, optarg);
            fflag++ ;
            break;
        case 'r':
        {
            char *end;
            datosAtmos->rugosidad = strtod(optarg, &end);
            if (end == optarg) {
                fprintf(stderr, "El parámetro -r no es un real\n");
                uso();
                exit(EXIT_FAILURE);
            }
            datosAtmos->tieneRugosidad = true;
            rflag++;
            break;
        }
        case 'R':
            datosAtmos->fich_topo = (char *)malloc(strlen(optarg)+1);
            if (datosAtmos->fich_topo == NULL) {
                fprintf(stderr, "Problema alocatando fich_topo\n");
                uso();
                exit(EXIT_FAILURE);
            }
            strcpy(datosAtmos->fich_topo,optarg);
            datosAtmos->tieneRugosidad = false;

            Rflag++;
            break;
        case 's':
            malla->fichPuntosSalida = (char *)malloc(strlen(optarg)+1);
            if (malla->fichPuntosSalida == NULL) {
                fprintf(stderr, "Problema alocatando fichPuntosSalida\n");
                uso();
                exit(EXIT_FAILURE);
            }
            strcpy(malla->fichPuntosSalida,optarg);
            malla->puntosSalida = true;

            break;
        case 'l':
            malla->tieneCapas = true;
            break;
        case 'i':
            *salidaInter = true;
            break;
        case 'n':
            *escribir = false;
            break;
        case '?':
            if ((optopt == 'c') || (optopt == 'f') || (optopt == 'r')
                    || (optopt == 'R') || (optopt == 's')) {
                fprintf (stderr, "El parametro -%c requiere un argumento adicional.\n", optopt);
            } else if (isprint(optopt)) {
                fprintf (stderr, "Opcion desconocida: `-%c'.\n", optopt);
            } else {
                fprintf (stderr,
                         "Caracter de opcion desconocido: `\\x%x'.\n",
                         optopt);
                exit(EXIT_FAILURE);
            }
        default:
            exit(EXIT_FAILURE);
        }
    }

    if (cflag == 0 || fflag == 0 || ((strcmp(file_format,"stations") == 0) && rflag == 0 && Rflag == 0)) {
        fprintf(stderr, "Error: Falta un parametro obligatorio\n\n");
        uso();
        exit(EXIT_FAILURE);
    }

    if (rflag == 1 && Rflag == 1) {
        fprintf(stderr, "Error: No se puede usar -r y -R simulataneamente\n\n");
        uso();
        exit(EXIT_FAILURE);
    }

    if (strcmp(file_format, "forecast") != 0 && strcmp(file_format, "stations") != 0 && strcmp(file_format, "interpolation")) {
        fprintf(stderr, "Error: %s: El parametro -f solo admite forecast | stations | interpolation \n\n", file_format);
        uso();
        exit(EXIT_FAILURE);
    }

}

/**
 * @brief Programa principal
 * @param argc Número de argumentos en la línea de comandos
 * @param argv Contenido de los argumetnos de la línea de comandos
 * @return exit status
 */
int main( int argc, char **argv )
{
    struct T_Topografia topo;
    struct T_NEstacion todas_est;
    struct T_Malla malla;
    struct T_DatosAtmos Datmos;
    struct VariableHARMONIE harm;
    char *infix = "out";
    bool salidaInter = false;
    bool escribir = true;

    // Leo la informacion de HARMONIE; se usa para construir el campo
    // inicial de viento
    Datmos.H = &harm;
    Datmos.tieneRugosidad = true;
    malla.puntosSalida = false;
    malla.puntosControl = false;
    malla.tieneCapas = false;

    char *fich_param, *file_format; // nombre del fichero de parametros

    fich_param = (char *) malloc(256 * sizeof(char));
    file_format = (char *) malloc(256 * sizeof(char));

    procesaArgumentos(argc, argv, fich_param, file_format, &Datmos, &malla, &salidaInter, &escribir);

    LeerParametros(fich_param, &Datmos);

    creaMalla(Datmos.fich_malla, &malla);

    leerViento(file_format, &Datmos, &todas_est, &topo, &malla);

    interpolacionViento(&Datmos, &malla, &topo, &todas_est, file_format);

    calculoViento(&malla, &Datmos);

    if (escribir){
        escribirResultados(&malla);
    }

    if (salidaInter){
        GrabarVientoInter("wind_inter.inp", &malla);
    }

    if (malla.tieneCapas){
        calculaCapas(&malla, &Datmos, &topo, &todas_est, file_format);
    }

    if (malla.puntosSalida){
        calculaPuntosSalida(&malla, &Datmos, &topo, &todas_est, file_format, infix);
    }

    libera_memoria(fich_param, file_format, &malla, &Datmos, &topo, &todas_est);

    return(EXIT_SUCCESS);
}

