#include "puntosSalida.h"

#include "localizador.h"
#include "viento.h"

#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

/**
 * @file puntosSalida.c
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
double**
leerPuntosSalida(const struct T_Malla *malla,
                 size_t *numeroPuntosOut)
{
    FILE *fp;
    char buffer[256];
    size_t line_number = 0;

    if(fp=fopen(malla->fichPuntosSalida, "r"),!fp){
        fprintf(stderr, "PuntosSalida:Problema con la apertura del fichero %s: %s\n",
                malla->fichPuntosSalida, strerror(errno));
        exit(EXIT_FAILURE);
    }

    size_t numeroPuntos=0;
    // Reserva memoria para las coordenadas de los nodos
    if (LeeLinea(fp, buffer, sizeof(buffer), &line_number) == 0){
        char *buf_dummy = buffer;
        char *p;
        numeroPuntos = strtoul(buf_dummy, &p, 10);
        if (buf_dummy == p){
            fprintf(stderr, "%s: not a decimal number\n", buf_dummy);
            fprintf(stderr, "%s:%zu\n",malla->fichPuntosSalida, line_number);
            exit(EXIT_FAILURE);
        }else{
            buf_dummy = p + 1;
        }
    } else {
        fprintf(stderr, "Error leyendo el número de puntos de salida\n");
        fprintf(stderr, "%s:%zu\n",malla->fichPuntosSalida, line_number);
        exit(EXIT_FAILURE);
    }


    double **ptosSalida = (double **)malloc(numeroPuntos*sizeof(double *));
    if (ptosSalida == NULL) {
        fprintf(stderr, "Ha habido un problema alocatando puntos salida\n");
        exit(EXIT_FAILURE);
    }

    for (size_t ptos = 0; ptos < numeroPuntos; ++ptos) {
        ptosSalida[ptos] = (double *)malloc(3*sizeof(double));
        if (ptosSalida[ptos]==NULL){
            fprintf(stderr, "Ha habido un problema alocatando puntos salida\n");
            exit(EXIT_FAILURE);
        }
    }

    size_t tipo=2;
    if (LeeLinea(fp, buffer, sizeof(buffer), &line_number) == 0) {

        if (strcmp(buffer,"terreno") == 0){
            tipo = 0;
        } else if (strcmp(buffer,"mar") == 0){
            tipo = 1;
        } else {
            fprintf(stderr, "Error, el tipo de viento no es ni 'terreno' ni 'mar'\n");
            fprintf(stderr, "%s:%zu\n",malla->fichPuntosSalida, line_number);
            exit(EXIT_FAILURE);
        }
    } else {
        fprintf(stderr, "Error leyendo el tipo de viento\n");
        fprintf(stderr, "%s:%zu\n",malla->fichPuntosSalida, line_number);
        exit(EXIT_FAILURE);
    }

    // Lectura de las coordenadas de los nodos
    {
        size_t counterPtos = 0;
        while (counterPtos < numeroPuntos){
            if (LeeLinea(fp, buffer, sizeof(buffer), &line_number) == 0) {

                parseLineAsListOfDoubles(3, buffer, malla->fichPuntosSalida, line_number, ptosSalida[counterPtos]);
                counterPtos++;
            } else {
                fprintf(stdout,"Problema leyendo las coordenadas del fichero de puntos de salida\n");
                fprintf(stderr, "%s:%zu\n",malla->fichPuntosSalida, line_number);
                exit(EXIT_FAILURE);
            }
        }
    }

    int iostat = fclose(fp);
    if (iostat != 0) {
        fprintf(stderr, "Problema al cerrar el archivo %s\n", malla->fichPuntosSalida);
        exit(EXIT_FAILURE);
    }


    if (tipo == 0){
        for (size_t ptosIdx = 0; ptosIdx < numeroPuntos; ++ptosIdx){
            ptosSalida[ptosIdx][2] += BuscarAltura(ptosSalida[ptosIdx], malla);
        }
    }

    *numeroPuntosOut = numeroPuntos;
    return(ptosSalida);

}

double**
leerPuntosControl(const struct T_Malla *malla,
                  size_t *numeroPuntosOut)
{
    FILE *fp;
    char buffer[256];
    size_t line_number = 0;

    if(fp=fopen(malla->fichPuntosControl, "r"),!fp){
        fprintf(stderr, "PuntosSalida:Problema con la apertura del fichero %s: %s\n",
                malla->fichPuntosControl, strerror(errno));
        exit(EXIT_FAILURE);
    }

    size_t numeroPuntos=0;
    // Reserva memoria para las coordenadas de los nodos
    if (LeeLinea(fp, buffer, sizeof(buffer), &line_number) == 0){
        char *buf_dummy = buffer;
        char *p;
        numeroPuntos = strtoul(buf_dummy, &p, 10);
        if (buf_dummy == p){
            fprintf(stderr, "%s: not a decimal number\n", buf_dummy);
            fprintf(stderr, "%s:%zu\n",malla->fichPuntosControl, line_number);
            exit(EXIT_FAILURE);
        }else{
            buf_dummy = p + 1;
        }
    } else {
        fprintf(stderr, "Error leyendo el número de puntos de control\n");
        fprintf(stderr, "%s:%zu\n",malla->fichPuntosSalida, line_number);
        exit(EXIT_FAILURE);
    }


    double **ptosControl = (double **)malloc(numeroPuntos*sizeof(double *));
    if (ptosControl == NULL) {
        fprintf(stderr, "Ha habido un problema alocatando puntos salida\n");
        exit(EXIT_FAILURE);
    }

    for (size_t ptos = 0; ptos < numeroPuntos; ++ptos) {
        ptosControl[ptos] = (double *)malloc(6*sizeof(double));
        if (ptosControl[ptos]==NULL){
            fprintf(stderr, "Ha habido un problema alocatando puntos salida\n");
            exit(EXIT_FAILURE);
        }
    }

    size_t tipo = 2;
    if (LeeLinea(fp, buffer, sizeof(buffer), &line_number) == 0) {

        if (strcmp(buffer,"modulo") == 0){
            tipo = 0;
        } else if (strcmp(buffer,"componentes") == 0){
            tipo = 1;
        } else {
            fprintf(stderr, "Error, el tipo de viento no es ni 'modulo' ni 'componentes'\n");
            fprintf(stderr, "%s:%zu\n",malla->fichPuntosControl, line_number);
            exit(EXIT_FAILURE);
        }
    } else {
        fprintf(stderr, "Error leyendo el tipo de viento\n");
        fprintf(stderr, "%s:%zu\n",malla->fichPuntosControl, line_number);
        exit(EXIT_FAILURE);
    }

    // Lectura de las coordenadas de los nodos
    {
        size_t longitud;

        if (tipo ==0){
            longitud = 5;
        } else {
            longitud = 6;
        }

        size_t counterPtos = 0;
        while (counterPtos < numeroPuntos){
            if (LeeLinea(fp, buffer, sizeof(buffer), &line_number) == 0) {
                parseLineAsListOfDoubles(longitud, buffer, malla->fichPuntosControl, line_number, ptosControl[counterPtos]);
                counterPtos++;
            } else {
                fprintf(stderr,"Problema leyendo los datos del fichero de puntos de control\n");
                fprintf(stderr, "%s:%zu\n",malla->fichPuntosControl, line_number);
                exit(EXIT_FAILURE);
            }
        }
    }

    int iostat = fclose(fp);
    if (iostat != 0) {
        fprintf(stderr, "Problema al cerrar el archivo %s\n", malla->fichPuntosSalida);
        exit(EXIT_FAILURE);
    }


    for (size_t ptosIdx = 0; ptosIdx < numeroPuntos; ++ptosIdx){
        ptosControl[ptosIdx][2] += BuscarAltura(ptosControl[ptosIdx], malla);
    }

    if (tipo==0){
        for (size_t ptosIdx = 0; ptosIdx < numeroPuntos; ++ptosIdx){
            double modulo = ptosControl[ptosIdx][3] ;
            double angulo = ptosControl[ptosIdx][4] ;
            const double rad = (270.0 - angulo) * 3.14159265358979323846 / 180.0;
            ptosControl[ptosIdx][3] = modulo *cos(rad);
            ptosControl[ptosIdx][4] = modulo * sin(rad);
            ptosControl[ptosIdx][5] = 0.0;

        }
    }

    *numeroPuntosOut = numeroPuntos;
    return(ptosControl);

}

void escribePuntosSalida(const double **ptosSalida, const struct T_Vector3D *vientoSalida,
                         const size_t numeroPuntos, const struct T_Malla *malla, const char *infix)
{
    int len = snprintf(NULL, 0, "%s%s_", malla->fichPuntosSalida, infix);
    char *fichero = (char *)malloc((len + 1)* sizeof(char));

    char *pch=strchr(malla->fichPuntosSalida,'.');
    char *fich1 = (char *)malloc((pch-malla->fichPuntosSalida+1)*sizeof(char));
    strncpy(fich1, malla->fichPuntosSalida, (pch-malla->fichPuntosSalida+1));
    fich1[pch-malla->fichPuntosSalida] = '\0';


    snprintf(fichero, len+1, "%s_%s%s", fich1, infix, pch);

    FILE *fp;

    fp=fopen(fichero, "w");
    if (!fp) {
        fprintf(stderr, "Problema al intentar abrir el archivo %s\n", fichero);
        exit(EXIT_FAILURE);
    }

    for (size_t ptosIdx=0; ptosIdx<numeroPuntos; ptosIdx++) {

        const double mod = sqrt(
                    (vientoSalida[ptosIdx].x*vientoSalida[ptosIdx].x) +
                    (vientoSalida[ptosIdx].y*vientoSalida[ptosIdx].y) +
                    (vientoSalida[ptosIdx].z*vientoSalida[ptosIdx].z)
                    );
        const double ang = ConvertWindCart2Met(vientoSalida[ptosIdx].x,
                                               vientoSalida[ptosIdx].y);
        fprintf(fp, "%f %f %f %f %f %f %f %f\n",
                ptosSalida[ptosIdx][0],
                ptosSalida[ptosIdx][1],
                ptosSalida[ptosIdx][2],
                vientoSalida[ptosIdx].x,
                vientoSalida[ptosIdx].y,
                vientoSalida[ptosIdx].z,
                mod,
                ang);
    }

    int iostat = fclose(fp);
    if (iostat != 0) {
        fprintf(stderr, "Problema al cerrar el archivo %s\n", fichero);
        exit(EXIT_FAILURE);
    }

    free(fichero);
    free(fich1);
}

void escribePuntosEstaciones(const double **ptosEstaciones, const struct T_Vector3D *vientoEstaciones,
                             const size_t numeroPuntos, const char *infix)
{
    int len = snprintf(NULL, 0, "estaciones_%s.dat", infix);
    char *fichero = (char *)malloc((len + 1)* sizeof(char));

    snprintf(fichero, len+1, "estaciones_%s.dat", infix);

    FILE *fp;

    fp=fopen(fichero, "w");
    if (!fp) {
        fprintf(stderr, "Problema al intentar abrir el archivo %s\n", fichero);
        exit(EXIT_FAILURE);
    }

    for (size_t ptosIdx=0; ptosIdx<numeroPuntos; ptosIdx++) {

        const double mod = sqrt(
                    (vientoEstaciones[ptosIdx].x*vientoEstaciones[ptosIdx].x) +
                    (vientoEstaciones[ptosIdx].y*vientoEstaciones[ptosIdx].y) +
                    (vientoEstaciones[ptosIdx].z*vientoEstaciones[ptosIdx].z)
                    );
        const double ang = ConvertWindCart2Met(vientoEstaciones[ptosIdx].x,
                                               vientoEstaciones[ptosIdx].y);
        fprintf(fp, "%f %f %f %f %f %f %f %f\n",
                ptosEstaciones[ptosIdx][0],
                ptosEstaciones[ptosIdx][1],
                ptosEstaciones[ptosIdx][2],
                vientoEstaciones[ptosIdx].x,
                vientoEstaciones[ptosIdx].y,
                vientoEstaciones[ptosIdx].z,
                mod,
                ang);
    }

    int iostat = fclose(fp);
    if (iostat != 0) {
        fprintf(stderr, "Problema al cerrar el archivo %s\n", fichero);
        exit(EXIT_FAILURE);
    }

    free(fichero);

}

void escribePuntosControl(const double **ptosControl, const struct T_Vector3D *vientoControl,
                          const size_t numeroPuntos, const struct T_Malla *malla, char *infix)
{
    int len = snprintf(NULL, 0, "%s_%s", malla->fichPuntosControl, infix);
    char *fichero = (char *)malloc((len + 1)* sizeof(char));

    char *pch=strchr(malla->fichPuntosControl,'.');
    char *fich1 = (char *)malloc((pch-malla->fichPuntosControl+1)*sizeof(char));
    strncpy(fich1, malla->fichPuntosControl, (pch-malla->fichPuntosControl+1));
    fich1[pch-malla->fichPuntosControl] = '\0';


    snprintf(fichero, len+1, "%s_%s%s", fich1, infix, pch);

    FILE *fp;

    fp=fopen(fichero, "w");
    if (!fp) {
        fprintf(stderr, "Problema al intentar abrir el archivo %s\n", fichero);
        exit(EXIT_FAILURE);
    }

    for (size_t ptosIdx=0; ptosIdx<numeroPuntos; ptosIdx++) {

        const double mod = sqrt(
                    (vientoControl[ptosIdx].x*vientoControl[ptosIdx].x) +
                    (vientoControl[ptosIdx].y*vientoControl[ptosIdx].y) +
                    (vientoControl[ptosIdx].z*vientoControl[ptosIdx].z)
                    );
        const double ang = ConvertWindCart2Met(vientoControl[ptosIdx].x,
                                               vientoControl[ptosIdx].y);

        fprintf(fp, "%f %f %f %f %f %f %f %f\n",
                ptosControl[ptosIdx][0],
                ptosControl[ptosIdx][1],
                ptosControl[ptosIdx][2],
                vientoControl[ptosIdx].x,
                vientoControl[ptosIdx].y,
                vientoControl[ptosIdx].z,
                mod,
                ang);
    }

    int iostat = fclose(fp);
    if (iostat != 0) {
        fprintf(stderr, "Problema al cerrar el archivo %s\n", fichero);
        exit(EXIT_FAILURE);
    }

    free(fichero);
    free(fich1);
}


void escribeInpPuntosSalida(const double **ptosSalida, const struct T_Vector3D *vientoSalida,
                            const size_t numeroPuntos, const struct T_Malla *malla, const char *infix)
{
    char *pch=strchr(malla->fichPuntosSalida,'.');
    char *fich1 = (char *)malloc((pch-malla->fichPuntosSalida+1)*sizeof(char));
    strncpy(fich1, malla->fichPuntosSalida, (pch-malla->fichPuntosSalida+1));
    fich1[pch-malla->fichPuntosSalida] = '\0';


    int len = snprintf(NULL, 0, "%s_%s%s",fich1, infix, ".inp");
    char *fichero = (char *)malloc((len + 1)* sizeof(char));
    snprintf(fichero, len+1, "%s_%s%s", fich1, infix, ".inp");

    FILE *fp;

    fp=fopen(fichero, "w");
    if (!fp) {
        fprintf(stderr, "Problema al intentar abrir el archivo %s\n", fichero);
        exit(EXIT_FAILURE);
    }

    fprintf(fp, "%zu %zu 3 0 0\n", numeroPuntos, numeroPuntos);

    for (size_t ptosIdx=0; ptosIdx<numeroPuntos; ptosIdx++) {
        fprintf(fp, "%zu %f %f %f\n", ptosIdx,
                ptosSalida[ptosIdx][0],
                ptosSalida[ptosIdx][1],
                ptosSalida[ptosIdx][2]);
    }

    for (size_t ptosIdx=0; ptosIdx<numeroPuntos; ptosIdx++) {
        fprintf(fp, "%zu %d pt %zu\n", ptosIdx,
                1,
                ptosIdx);
    }

    fprintf(fp, "1 3\nV, m/s\n");

    for (size_t ptosIdx=0; ptosIdx<numeroPuntos; ptosIdx++) {
        fprintf(fp, "%zu %f %f %f\n", ptosIdx,
                vientoSalida[ptosIdx].x,
                vientoSalida[ptosIdx].y,
                vientoSalida[ptosIdx].z);
    }

    int iostat = fclose(fp);
    if (iostat != 0) {
        fprintf(stderr, "Problema al cerrar el archivo %s\n", fichero);
        exit(EXIT_FAILURE);
    }

    free(fichero);
    free(fich1);

}


size_t *
getPuntosSalidaTetra(const size_t numeroPuntos, const struct T_Malla* malla, const double **ptosSalida)
{
    size_t *tetra = (size_t *)malloc(numeroPuntos*sizeof(size_t));
    if (tetra == NULL){
        fprintf(stderr, "Ha habido un problema alocatando tetra\n");
        exit(EXIT_FAILURE);
    }

    for (size_t ptosIdx = 0; ptosIdx < numeroPuntos; ++ptosIdx){
        tetra[ptosIdx] = getTetrahedra(ptosSalida[ptosIdx][0],
                ptosSalida[ptosIdx][1],
                ptosSalida[ptosIdx][2],
                (const double**)(malla->nodo), (const size_t**)malla->tetra, malla->Num_tetra, &malla->situation);

        if ( tetra[ptosIdx] == malla->Num_tetra+1) {
            fprintf(stderr, "No se ha encontrado tetraedro\n");
            exit(EXIT_FAILURE);
        }
    }

    return tetra;
}

struct T_Vector3D*
        getVientoSalida(const size_t numeroPuntos, const struct T_NEstacion *todas_est,
                        const struct T_DatosAtmos *Datmos, const struct T_Malla *malla,
                        const struct T_Topografia *topo, const char *file_format,
                        const double **ptosSalida, const size_t *tetra)
{
    struct T_Vector3D *vientoSalida = (struct T_Vector3D *)malloc(numeroPuntos*sizeof(struct T_Vector3D));
    if (vientoSalida == NULL){
        fprintf(stderr, "Ha habido un problema alocatando viento salida\n");
        exit(EXIT_FAILURE);
    }
    struct T_Vector3D vini;
    for (size_t ptosIdx = 0; ptosIdx < numeroPuntos; ++ptosIdx){
        vientoSalida[ptosIdx] = VientoEnPunto(malla,
                                              topo, Datmos, todas_est,
                                              tetra[ptosIdx],
                                              ptosSalida[ptosIdx][0],
                ptosSalida[ptosIdx][1],
                ptosSalida[ptosIdx][2],
                BuscarAltura(ptosSalida[ptosIdx], malla),
                file_format,&vini);
    }

    return vientoSalida;
}

void
calculaPuntosSalida(const struct T_Malla *malla, const struct T_DatosAtmos *Datmos,
                    const struct T_Topografia *topo, const struct T_NEstacion *todas_est, const char *file_format, const char *infix)
{

    size_t numeroPuntos;
    double **ptosSalida = leerPuntosSalida(malla, &numeroPuntos);


    size_t *tetra = getPuntosSalidaTetra(numeroPuntos, malla, (const double**)ptosSalida);

    struct T_Vector3D *vientoSalida = getVientoSalida(numeroPuntos, todas_est, Datmos,
                                                      malla, topo, file_format, (const double**)ptosSalida, tetra);

    escribePuntosSalida((const double**)ptosSalida, vientoSalida, numeroPuntos, malla, infix);
    escribeInpPuntosSalida((const double **)ptosSalida, vientoSalida, numeroPuntos, malla, infix);

    free(vientoSalida);
    free(tetra);
    for (size_t ptos = 0; ptos < numeroPuntos; ++ptos) {
        free(ptosSalida[ptos]);
    }
    free(ptosSalida);

}

void
escribeEstaciones(const struct T_Malla *malla, const struct T_DatosAtmos *Datmos,
                  const struct T_Topografia *topo, const struct T_NEstacion *todas_est, const char *file_format,
                  const char *infix)
{

    size_t numeroPuntos = todas_est->NEST;
    double **ptosEstaciones = (double **)malloc(numeroPuntos * sizeof(double *));
    if (ptosEstaciones == NULL){
        fprintf(stderr, "Ha ocurrido un error alocatando ptosEstaciones\n");
        exit(EXIT_FAILURE);
    }

    for (size_t j = 0; j < numeroPuntos; ++j)
    {
        ptosEstaciones[j] = (double *)malloc(3*sizeof(double));
        if (ptosEstaciones[j] == NULL){
            fprintf(stderr, "Ha ocurrido un error alocatando ptosEstaciones\n");
            exit(EXIT_FAILURE);
        }
        ptosEstaciones[j][0] = todas_est->N[j]->x;
        ptosEstaciones[j][1] = todas_est->N[j]->y;
        const double punto[3] = {
            todas_est->N[j]->x,
            todas_est->N[j]->y,
            todas_est->N[j]->z};
        const double alturaTerr = BuscarAltura(punto,malla);
        ptosEstaciones[j][2] = todas_est->N[j]->z_terr + alturaTerr;

    }


    size_t *tetra = getPuntosSalidaTetra(numeroPuntos, malla, (const double**)ptosEstaciones);

    struct T_Vector3D *vientoEstaciones = getVientoSalida(numeroPuntos, todas_est, Datmos,
                                                          malla, topo, file_format, (const double**)ptosEstaciones, tetra);


    escribePuntosEstaciones((const double**)ptosEstaciones, vientoEstaciones, numeroPuntos, infix);


    free(vientoEstaciones);
    free(tetra);
    for (size_t ptos = 0; ptos < numeroPuntos; ++ptos) {
        free(ptosEstaciones[ptos]);
    }
    free(ptosEstaciones);

}
