#ifndef PUNTOSSALIDA_H
#define PUNTOSSALIDA_H

#include "lectura.h"

/**
 * @file puntosSalida.h
 * @brief Fichero con todas las funciones relativas a la lectura, calculo y
 * escritura de los puntos de salida
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @brief Función que lee  el fichero de los puntos de salida
 * @param[in] malla Vector que almacena la malla y el fichero de puntos de salida
 * @param[out] numeroPuntosOut Número de puntos de salida
 * @return Vector con los puntos de salida
 */
double**
leerPuntosSalida(const struct T_Malla *malla,
                 size_t *numeroPuntosOut);

/**
 * @brief Función que lee  el fichero de los puntos de control
 * @param[in] malla Vector que almacena la malla y el fichero de puntos de control
 * @param[out] numeroPuntosOut Número de puntos de control
 * @return Vector con los puntos de control
 */
double**
leerPuntosControl(const struct T_Malla *malla,
                 size_t *numeroPuntosOut);

/**
 * @brief Función que una vez calculado el viento en los puntos de salida los escribe en un fichero
 * @param[in] ptosSalida Vector con los puntos de salida
 * @param[in] vientoSalida Vector con el viento en los puntos de salida
 * @param[in] numeroPuntos Número de puntos de salida
 * @param[in] malla Variable con la estructura de la malla
 * @param[in] infix Cadena de carácteres para distinguir los ficheros de salida
 */
void escribePuntosSalida(const double **ptosSalida, const struct T_Vector3D *vientoSalida,
                         const size_t numeroPuntos, const struct T_Malla *malla, const char *infix);

/**
 * @brief Función que una vez calculado el viento en los puntos de control los escribe en un fichero
 * @param[in] ptosControl  Vector con los puntos de control
 * @param[in] vientoControl Vector con el viento en los puntos de control
 * @param[in] numeroPuntos Número de puntos de control
 * @param[in] malla Variable con la estructura de la malla
 * @param[in] infix Cadena de carácteres para distinguir los ficheros de salida
 */
void escribePuntosControl(const double **ptosControl, const struct T_Vector3D *vientoControl,
                         const size_t numeroPuntos, const struct T_Malla *malla, char *infix);

/**
 * @brief Función que una vez calculado el viento en los puntos de control los escribe en un fichero
 * @param[in] ptosEstaciones Vector con las estaciones
 * @param[in] vientoEstaciones Vector con el viento en las estaciones
 * @param[in] numeroPuntos Número de estaciones
 * @param[in] infix Cadena de carácteres para distinguir los ficheros de salida
 */
void escribePuntosEstaciones(const double **ptosEstaciones, const struct T_Vector3D *vientoEstaciones,
                         const size_t numeroPuntos, const char *infix);



/**
 * @brief Función que una vez calculado el viento en los puntos de salida los escribe en un fichero INP
 * @param[in] ptosSalida Vector con los puntos de salida
 * @param[in] vientoSalida Vector con el viento en los puntos de salida
 * @param[in] numeroPuntos Número de puntos de salida
 * @param[in] malla Variable con la estructura de la malla
 * @param[in] infix Cadena de carácteres para distinguir los ficheros de salida
 */
void escribeInpPuntosSalida(const double **ptosSalida, const struct T_Vector3D *vientoSalida,
                            const size_t numeroPuntos, const struct T_Malla *malla, const char *infix);

/**
 * @brief Función que encuentra los tetraedros en los que se encuentran los puntos de salida
 * @param[in] numeroPuntos Número de puntos de salida
 * @param[in] malla Variable con la estructura de malla
 * @param[in] ptosSalida Vector con los puntos de salida
 * @return Vector con los tetraedros en los que se encuentran los puntos de control
 */
size_t *
getPuntosSalidaTetra(const size_t numeroPuntos, const struct T_Malla *malla, const double **ptosSalida);

/**
 * @brief Función que calcula el viento en los puntos
 * @param[in] numeroPuntos Número de puntos en los que calcular el viento
 * @param[in] todas_est Variable con los datos de las estaciones
 * @param[in] Datmos Variable con los datos atmosféricos
 * @param[in] malla Variable con la estructura de la malla
 * @param[in] topo Variable con la longitud de rugosidad
 * @param[in] file_format Tipo de problema (stations, forecast interpolation)
 * @param[in] ptosSalida Vector con los puntos
 * @param[in] tetra Vector con los tetraedros en los que se encuentra cada punto
 * @return Vector con el campo de viento en cada punto
 */
struct T_Vector3D*
        getVientoSalida(const size_t numeroPuntos, const struct T_NEstacion *todas_est,
                        const struct T_DatosAtmos *Datmos, const struct T_Malla *malla,
                        const struct T_Topografia *topo, const char *file_format,
                        const double **ptosSalida, const size_t *tetra);

/**
 * @brief Función que dado un fichero con los puntos de salida, los lee, calcula su viento y escribe el fichero de resultados
 * @param[in] malla Variable con la estructura de la malla
 * @param[in] Datmos Variable con los datos atmosféricos
 * @param[in] topo Variable con las longitudes de rugosidad
 * @param[in] todas_est Variables con los datos de las estaciones
 * @param[in] file_format Tipo de problema (Stations, forecast interpolation)
 * @param[in] infix Cadena que se añade al nombre del fichero para indicar que es de salida
 */
void calculaPuntosSalida(const struct T_Malla *malla, const struct T_DatosAtmos *Datmos,
                    const struct T_Topografia *topo, const struct T_NEstacion *todas_est,
                         const char *file_format, const char *infix);

/**
 * @brief Función que escribe el campo de viento en las estaciones
 * @param[in] malla Variable con la estructura de la malla
 * @param[in] Datmos Variable con los datos atmosféricos
 * @param[in] topo Variable con las longitudes de rugosidad
 * @param[in] todas_est Variables con los datos de las estaciones
 * @param[in] file_format Tipo de problema (Stations, forecast interpolation)
 * @param[in] infix Cadena que se añade al nombre del fichero para indicar que es de salida
 */
void
escribeEstaciones(const struct T_Malla *malla, const struct T_DatosAtmos *Datmos,
                    const struct T_Topografia *topo, const struct T_NEstacion *todas_est,
                  const char *file_format,const  char *infix);


#endif // PUNTOSSALIDA_H
