#include "topografia.h"

#include "comparisons.h"
#include "debugInfo.h"
#include "localizador.h"

#include <math.h>
#include <stdlib.h>
#include <string.h>

/**
 * @file topografia.c
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

void uso_topo()
{
    fprintf(stdout,"\nFormato del fichero de topografia: \
            Si una linea comienza por el caracter '#' se ignora (comentario). \
            El fichero comienza con seis lineas en este orden: \
            \tminimo de la coordenada x (xmin) \
            \tmaximo de la coordenada x (xmax) \
            \tminimo de la coordenada y (ymin) \
            \tmaximo de la coordenada y (ymax) \
            \tincremento de x (deltax) \
            \tincremento de y (deltay)\
            \
            A continuacion viene la informacion topografica, en lineas con el \
            siguiente formato: \
            \tx y z rugosidad\n\n");
}

void leer_topo(const char *fich, struct T_Topografia *topo)
{
    static char *nfun="leer_topo()";
    FILE *f;
    double **matriz, **z0, x, y, z, rugosidad;
    double xmin, xmax, ymin, ymax, deltax, deltay;
    double aux[6];
    size_t nx, ny, cx, cy, i, line_number=0;
    size_t Nl, l=0;
    char buf[256];

    if ( (f = fopen(fich,"r")) == NULL ){
        fprintf(stderr, "%s\n", fich);
        fprintf(stderr, "Error en %s:\n", nfun);
        exit(EXIT_FAILURE);
    }
    // lee la discretizacion de la topografia
    for(i=0; i<6; i++){
        if(LeeLinea(f, buf, sizeof(buf), &line_number)!=0){
            fprintf(stderr,"leer_topo: No puedo leer los datos generales de la topografia\n");
            fprintf(stderr, "%s:%zu\n", fich, line_number);
            uso_topo();
            exit(EXIT_FAILURE);
        }
        char* endPtr;
        aux[i] = strtod(buf, &endPtr);
        l++;
    }

    xmin = aux[0];   xmax = aux[1];
    ymin = aux[2];   ymax = aux[3];
    deltax  = aux[4]; deltay = aux[5];
    l = 0;

    // Calcula la dimension que tendra la matriz de la topografia
    nx = topo->nx = (size_t)((xmax - xmin) / deltax + 1.0);
    ny = topo->ny =(size_t)((ymax - ymin) / deltay + 1.0);
    // Numero de lineas a leer del fichero
    Nl = (unsigned long int) nx * ny;

    // Reserva espacio para la matriz de la topografia
    if ((matriz = (double **)malloc(sizeof(double*) * nx)) == NULL){
        fprintf(stderr, "leer_topo: no puedo reservar memoria para matriz\n");
        exit(EXIT_FAILURE);
    }
    for (i=0; i< nx; i++) {
        if ((matriz[i] = (double *)malloc(sizeof(double) * ny)) == NULL){
            fprintf(stderr, "leer_topo: no puedo reservar memoria para matriz[%zu]\n", i);
            exit(EXIT_FAILURE);
        }
    }

    // Reserva espacio para la matriz de la rugosidad
    if ((z0 = (double **)malloc(sizeof(double*) * nx)) == NULL){
        fprintf(stderr, "leer_topo: no puedo reservar memoria para z0\n");
        exit(EXIT_FAILURE);
    }
    for (i=0; i< nx; i++) {
        if ((z0[i] = (double *)malloc(sizeof(double) * ny)) == NULL){
            fprintf(stderr, "leer_topo: no puedo reservar memoria para z0[%zu]\n", i);
            exit(EXIT_FAILURE);
        }
    }

    while (l < Nl){
        if (LeeLinea(f, buf, sizeof(buf), &line_number) == 0){

            double a[4];
            parseLineAsListOfDoubles(4, buf, fich, line_number, a);

            x = a[0];
            y = a[1];
            z = a[2];
            rugosidad = a[3];

            l++;

            cx = (x - xmin) / deltax;
            cy = (y - ymin) / deltay;

            matriz[cx][cy] = z;
            z0[cx][cy] = rugosidad;
        } else {
            fprintf(stderr, "Error leyendo el fichero de topografia\n");
            fprintf(stderr,"%s:%zu", fich, line_number);
            uso_topo();
            exit(EXIT_FAILURE);
        }
    }

    fclose(f);

    topo->z = matriz;
    topo->z0 = z0;
    topo->deltax = deltax;
    topo->deltay = deltay;
    topo->xmin = xmin; topo->xmax = xmax;
    topo->ymin = ymin; topo->ymax = ymax;
}


/**
 * Dado un punto (X,Y) en coordenadas de la topografia original
 * calcula la topografia en ese punto a traves de una interpolacion
 * de Lagrange de cuatro puntos. Esta funcion solo es valida para
 * rectangulos de lados paralelos a los ejes, como en nuestro caso
 */
double InterpolaTopo(const double *x, const double *y, const struct T_Topografia *topo)
{
    int x1, y1, x2, y2;
    double N1, N2, N3, N4, D1, D2, D3, D4;
    double z0;

    // transforma a coordenadas de la matriz de topografia
    double X = ((*x) - topo->xmin)/topo->deltax;
    double Y = ((*y) - topo->ymin)/topo->deltay;

    // calcula entre que puntos hay que interpolar
    x1 = (int) floor((double) X );  y1 = (int) floor((double) Y );
    x2 = (int) ceil((double) X );   y2 = (int) ceil((double) Y );

    // Si el punto a interpolar coincide con alguno de los puntos
    // dato, ceil y floor devuelven el mismo resultado, y eso
    // provocoria problemas
    if (x1 == x2) x2++;
    if (y1 == y2) y2++;

    D1 = (double) ((x2 - x1) * (y2 - y1));
    D2 = -D1;
    D3 = D1;
    D4 = D1;

    N1 = (double) (((x2 - X) * (y2 - Y)) / D1);
    N2 = (double) (((x1 - X) * (y2 - Y)) / D2);
    N3 = (double) (((x1 - X) * (y1 - Y)) / D3);
    N4 = (double) (((X - x2) * (y1 - Y)) / D4);


    z0 = (double) (N1 * topo->z0[x1][y1] + N2 * topo->z0[x2][y1] +
                   N3 * topo->z0[x2][y2] + N4 * topo->z0[x1][y2]);


    return(z0);
}



void LeerEstacion(FILE *fp, struct T_Estacion *est, const size_t tipo, const char *fich, size_t *line_number)
{
    double values[5];
    char buffer[512];

    if (LeeLinea(fp, buffer, sizeof(buffer), line_number) == 0) {
        parseLineAsListOfDoubles(5, buffer, fich, *line_number, values);
    } else {
        fprintf(stderr, "problem reading wind file\n");
        fprintf(stderr, "%s:%zu\n", fich, *line_number);
        exit(EXIT_FAILURE);
    }

    est->x = values[0];
    est->y = values[1];
    est->z = values[2];
    if (tipo == 0){
        est->modulo = values[3];
        est->angulo = values[4];
        const double rad = (270.0 - values[4]) * 3.14159265358979323846 / 180.0;
        est->vx = values[3] * cos(rad);
        est->vy = values[3] * sin(rad);
        est->vz = 0.0;
    } else {
        est->vx = values[3];
        est->vy = values[4];
        est->vz = 0;
    }


}

/**
 * La altura de la estacion sobre el terreno es calculada con
 * una interpolacion de sus coordenadas x,y sobre la topografia
 * y sumandole la altura estandar del anemometro (normalmente 10m)
 */
void correccionAlturaEstaciones(const struct T_DatosAtmos *Datmos, const double alturaAnemometro,
                                struct T_NEstacion *Est, const double* altEstaciones,
                                const size_t i, const struct T_Topografia *topo)
{
    struct T_Vector3D Vh;

    double z0;
    if (Datmos->tieneRugosidad) {
        z0 = Datmos->rugosidad;
    } else {
        z0 = InterpolaTopo(&(Est->N[i]->x), &(Est->N[i]->y), topo);
    }

    Vh.x = Est->N[i]->vx_orig;
    Vh.y = Est->N[i]->vy_orig;
    Vh.z = Est->N[i]->vz_orig;

    const double altEst = *altEstaciones;

    InterpolacionVertical( altEst, z0, Datmos , alturaAnemometro, &Vh);
    Est->N[i]->vx = Vh.x;
    Est->N[i]->vy = Vh.y;
    Est->N[i]->vz = 0.0;

}

void leer_interpolacion(const char *fich_viento, struct T_Malla *malla)
{
    size_t line_number = 0;

    FILE *fp = fopen(fich_viento, "r");
    if (fp == NULL){
        fprintf(stderr, "leer_interpolacion: error al abrir el fichero %s\n", fich_viento);
        exit(EXIT_FAILURE);
    }

    size_t counterPtos = 0;
    while ((counterPtos < malla->Num_tetra)){
        char buffer[256];
        if (LeeLinea(fp, buffer, sizeof(buffer), &line_number) == 0) {
            parseLineAsListOfDoubles(3, buffer, fich_viento, line_number, malla->vini[counterPtos]);
            counterPtos++;
        } else {
            fprintf(stderr, "Error leyendo fichero interpolación\n");
            fprintf(stderr, "%s:%zu\n", fich_viento, line_number);
            exit(EXIT_FAILURE);
        }

    }

    int iostat = fclose(fp);
    if (iostat != 0) {
        fprintf(stderr, "Problema al cerrar el archivo %s\n", fich_viento);
        exit(EXIT_FAILURE);
    }


}

size_t LeerLasEstaciones(const char *fich, const double* altEstaciones,
                         struct T_NEstacion *Est, const struct T_Malla *malla)
{

    FILE *fp;
    size_t numeroEstaciones=0, line_number = 0;
    char buffer[512];

    if ( (fp = fopen(fich,"r")) == NULL ){
        fprintf(stderr, "LeerEstacion:No se pudo abrir el fichero");
        fprintf(stderr, "Nombre del fichero %s\n", fich);
        exit(EXIT_FAILURE);
    }

    // Reserva memoria para las coordenadas de los nodos
    if (LeeLinea(fp, buffer, sizeof(buffer), &line_number) == 0) {
        char *buf_dummy = buffer;
        char *p;
        numeroEstaciones = strtoul(buf_dummy, &p, 10);
        if (buf_dummy == p) {
            fprintf(stderr, "%s: not a decimal number\n", buf_dummy);
            fprintf(stderr, "%s:%zu\n", fich, line_number);
            exit(EXIT_FAILURE);
        } else {
            buf_dummy = p + 1;
        }
    }

    size_t tipo=0;
    if (LeeLinea(fp, buffer, sizeof(buffer), &line_number) == 0) {

        if (strcmp(buffer,"modulo") == 0){
            tipo = 0;
        } else if (strcmp(buffer,"componentes") == 0){
            tipo = 1;
        } else {
            fprintf(stderr, "Error, el tipo de viento no es ni 'componentes' ni 'modulo'\n");
            fprintf(stderr, "%s:%zu\n", fich, line_number);
            exit(EXIT_FAILURE);
        }
    }

    size_t i;

    Est->NEST = numeroEstaciones;
    Est->noMismaAltura = false;
    Est->Nfo = 0;
    Est->N = (struct T_Estacion **) malloc(numeroEstaciones * sizeof(struct T_Estacion *));
    for (i = 0; i < numeroEstaciones; i++){
        Est->N[i] = (struct T_Estacion *) malloc(sizeof(struct T_Estacion));
        LeerEstacion(fp, Est->N[i], tipo, fich, &line_number);
        Est->N[i]->vx_orig = Est->N[i]->vx;
        Est->N[i]->vy_orig = Est->N[i]->vy;
        Est->N[i]->vz_orig = Est->N[i]->vz;
        double alturaAnemometro = Est->N[i]->z;
        Est->N[i]->z_terr = Est->N[i]->z;
        if (!equal(&alturaAnemometro, altEstaciones))
        {
            Est->noMismaAltura = true;
            Est->N[i]->noMismaAltura = true;
        }
        else
        {
            Est->N[i]->noMismaAltura = false;
        }

        // Calcula la altura de la estacion interpolando sus coordenadas
        // sobre la topografia; ademas suma la altura estandar del anemometro
        // sobre el suelo (normalmente 10m, el valor concreto esta almacenado
        // en datmos->altest
        double punto[3];
        punto[0] = Est->N[i]->x;
        punto[1] = Est->N[i]->y;
        punto[2] = Est->N[i]->z;
        double alturaTerr = BuscarAltura(punto,malla);
        Est->N[i]->z = (*altEstaciones) +
                alturaTerr;

    }

    fclose(fp);
    return(i);
}

/**
 * Se pondera mediante epsilon el inverso de la distancia al cuadrado
 * y la diferencia de alturas
 */
void InterpolacionHorizontal(const double *x, const double *y, const double *z_terr,
                             const double *epsilon,
                             const double* alturaEstaciones,
                             const struct T_NEstacion *est,
                             struct T_Vector3D *V)
{
    int icontrol, jcontrol;
    double *DD, *DH;
    double dden=0.0, valor=0.0, dnumx=0.0, dnumy=0.0;
    char *nfun = "InterpolacionHorizontal";

    V->z = 0.0;
    icontrol = jcontrol = -1;
    DD = (double *)malloc(sizeof(double) * est->NEST);
    DH = (double *)malloc(sizeof(double) * est->NEST);

    { size_t k=0;
        for (size_t i=0; i<est->NEST; i++) {
            k++ ;
            DD[i] = sqrt( ((*x) - est->N[i]->x) * ((*x) - est->N[i]->x) +
                          ((*y) - est->N[i]->y) * ((*y) - est->N[i]->y) ) ;
            if ( equalToZero(&DD[i]) ) {

                icontrol = i;
            }
            //            double punto[3];
            //            punto[0] = *x;
            //            punto[1] = *y;

            DH[i] = fabs(est->N[i]->z -
                         (
                             //                         BuscarAltura(punto, malla)
                             (*z_terr)
                             + (*alturaEstaciones)) ) ;
            if (equalToZero(&DH[i]) ){

                jcontrol = i;
            }
        }

        if (k < 1) {
            fprintf(stderr, "%s: Error: No hay estaciones para calcular el viento.\n", nfun);
            fprintf(stderr, "Todas las estaciones son de referencia (codicocalculo)\n");
            exit(EXIT_FAILURE);
        }
    }

    if (icontrol < 0) {
        for(size_t k=0; k < est->NEST; k++) {
            valor = 1.0 / (DD[k] * DD[k]);
            dden += valor;
            dnumx = dnumx + valor * est->N[k]->vx;
            dnumy = dnumy + valor * est->N[k]->vy;
        }

        V->x = dnumx * (*epsilon) / dden;
        V->y = dnumy * (*epsilon) / dden;

        if (jcontrol < 0) {
            dden = dnumx = dnumy = 0.0;
            for(size_t k=0; k < est->NEST; k++) {

                valor = 1.0 / DH[k];
                dden += valor;
                dnumx = dnumx + valor * est->N[k]->vx;
                dnumy = dnumy + valor * est->N[k]->vy;
            }
            V->x += (dnumx * (1.0 - (*epsilon)) / dden);
            V->y += (dnumy * (1.0 - (*epsilon)) / dden);
        } else {
            {
                V->x = V->x + (1.0 - (*epsilon)) * est->N[jcontrol]->vx;
                V->y = V->y + (1.0 - (*epsilon)) * est->N[jcontrol]->vy;
            }
        }
    } else {
        {
            V->x = est->N[icontrol]->vx;
            V->y = est->N[icontrol]->vy;
        }
    }

    free(DD);
    free(DH);
}



double CalculaInversaLongitudMO(const int estabilidad, const double *z0)
{
    double a, b;

    switch (estabilidad) {
    case -3:
        // Clase A: extremadamente inestable
        a = -0.0875;
        b = -0.1029;
        break;
    case -2:
        // Clase B: moderadamente inestable
        a = -0.03849;
        b = -0.1714;
        break;
    case -1:
        // Clase C: ligeramente inestable
        a = -0.00807;
        b = -0.3049;
        break;
    case 0:
        // Clase D: neutra
        a = b = 0.0;
        break;
    case 1:
        // Clase E: ligeramente estable
        a = 0.00807;
        b = -0.3049;
        break;
    case 2:
        // Clase F: moderadamente estable
        a = 0.03849;
        b = -0.1714;
        break;
    case 3:
        // Clase G: extremadamente estable
        a = 0.0875;
        b = -0.1029;
        break;
    default:
        fprintf(stderr, "FL: La clase de estabilidad esta fuera de rango\n");
        exit(EXIT_FAILURE);
    }
    a = a * exp(b * log(*z0));
    return(a);
}// FL


double calculaFI(const int pasquill, const double* z0, const double* z)
{
    double dl, theta;

    if (pasquill == 0) {
        return(0.0);
    }
    dl = CalculaInversaLongitudMO(pasquill, z0);
    if (pasquill > 0) {
        return(-5.0 * (*z) * dl);
    }

    theta = 1.0 - 16.0 * (*z) * dl;
    theta = exp(0.25 * log(theta));

    dl = log( (((theta * theta) + 1.0) * (theta + 1.0) * (theta + 1.0)) / 8.0 );
    dl =  dl - (2 * atan(theta)) + (PI/2.0);

    return(dl);
}


void InterpolacionVertical(const double alturaPunto, const double alturaRugosidad,
                           const struct T_DatosAtmos *datmos,
                           const double alturaEstacion, struct T_Vector3D *V)
{
    double Phi_m, modvfric, Zpbl, h, rho;
    struct T_Vector3D Vfric;


    Phi_m = calculaFI(datmos->pasquill, &alturaRugosidad, &alturaEstacion);
    double aux = datmos->dk / (log(alturaEstacion / alturaRugosidad) - Phi_m);
    Vfric.x = V->x * aux;
    Vfric.y = V->y * aux;
    Vfric.z = V->z * aux;
    modvfric = sqrt(Vfric.x*Vfric.x + Vfric.y*Vfric.y + Vfric.z*Vfric.z);

    Zpbl = datmos->gamma * modvfric / datmos->f;
    if (datmos->pasquill <= 0) { // atmosfera inestable o neutra
        h = Zpbl;
    } else {                     // atmosfera estable
        h = datmos->gamma2 * sqrt( modvfric *
                                   (1.0/CalculaInversaLongitudMO(datmos->pasquill, &alturaRugosidad)) /
                                   datmos->f );
    }
    double Zsl = h / 10.0;


    DEBUG_INFO(
                fprintf(stdout, "z0=%010.6f \tZsl=%010.6f \th=%010.6f \tZpbl=%010.6f\n",
                       alturaRugosidad, Zsl, h, Zpbl);
            );


    if ((alturaPunto >= alturaRugosidad) && (alturaPunto <= Zsl)){
        Phi_m = calculaFI(datmos->pasquill, &alturaRugosidad, &alturaPunto);
        aux = (log(alturaPunto / alturaRugosidad) - Phi_m) / datmos->dk;
        V->x = Vfric.x * aux;
        V->y = Vfric.y * aux;
        V->z = 0.0;
    }

    else if ((alturaPunto > Zsl) && (alturaPunto <= Zpbl)){
        Phi_m = calculaFI(datmos->pasquill, &alturaRugosidad, &Zsl);
        aux = (log(Zsl / alturaRugosidad) - Phi_m) / datmos->dk;
        V->x = Vfric.x * aux;
        V->y = Vfric.y * aux;
        aux = (alturaPunto - Zsl) / (Zpbl - Zsl);
        rho = 1.0 - (aux * aux) * (3.0 - 2.0 * aux);
        V->x = rho * V->x + (1.0 - rho) * datmos->vg.x;
        V->y = rho * V->y + (1.0 - rho) * datmos->vg.y;
        V->z = 0.0;

    }
    else
    {

        V->x = datmos->vg.x;
        V->y = datmos->vg.y;
        V->z = 0.0;
    }

}

struct T_Vector3D InterpolacionViento(const double x, const double y, const double z,
                                      const double z_terr,
                                      const struct T_DatosAtmos *datmos,
                                      const struct T_NEstacion *nest,
                                      const struct T_Topografia *topo)
{
    struct T_Vector3D V;
    double zm = datmos->altest;
    double z0;
    if (datmos->tieneRugosidad) {
        z0 = datmos->rugosidad;
    } else {
        z0 = InterpolaTopo(&x, &y, topo);
    }
    //    double punto[3];
    //    punto[0] = x;
    //    punto[1] = y;
    //    punto[2] = z;
    double zAux = z - z_terr; //BuscarAltura(punto, malla);

    if (zAux < z0){
        V.x = V.y = V.z = 0.0;
        return(V);
    }

    InterpolacionHorizontal(&x, &y, &z_terr, &datmos->epsilon, &datmos->altest, nest, &V);

    InterpolacionVertical(zAux, z0, datmos, zm, &V);

    return(V);
}

void getNormalVector(const double *p1, const double *p2, const double *p3, double *n)
{


    double V1[3] = {p2[0]-p1[0], p2[1]-p1[1], p2[2]-p1[2]};
    double V2[3] = {p3[0]-p1[0], p3[1]-p1[1], p3[2]-p1[2]};


    n[0] = (V1[1]*V2[2])-(V1[2]*V2[1]);
    n[1] = (V1[2]*V2[0])-(V1[0]*V2[2]);
    n[2] = (V1[0]*V2[1])-(V1[1]*V2[0]);

    double modulus = sqrt((n[0]*n[0])+(n[1]*n[1])+(n[2]*n[2]));

    n[0] = n[0]/modulus;
    n[1] = n[1]/modulus;
    n[2] = n[2]/modulus;
}




double getHeight(const double *punto1, const double *punto2, const double *punto3, const double *punto)
{


    double *n = (double *)malloc(3*sizeof(double));
    getNormalVector(punto1, punto2, punto3, n);

    const double x0 = punto1[0];
    const double y0 = punto1[1];
    const double z0 = punto1[2];

    const double x = punto[0];
    const double y = punto[1];

    const double term1 = n[0]*(x-x0);
    const double term2 = n[1]*(y-y0);
    const double sumterm = term1 + term2;
    const double divterm = sumterm/n[2];

    const double height = z0-divterm;

    DEBUG_INFO(
                const double h_max = fmax(punto1[2], fmax(punto2[2], punto3[2]));
    const double h_min = fmin(punto1[2], fmin(punto2[2], punto3[2]));

    if (lessThan(&height,&h_min) || greatThan(&height,&h_max)) {
        fprintf(stdout, "comprobacion fallada\n");
        fprintf(stdout, "punto: %f %f %f \n",
                punto[0], punto[1], height);

        fprintf(stdout, "punto1: %f %f %f \n",
                punto1[0], punto1[1], punto1[2]);

        fprintf(stdout, "punto2: %f %f %f \n",
                punto2[0], punto2[1], punto2[2]);

        fprintf(stdout, "punto3: %f %f %f \n",
                punto3[0], punto3[1], punto3[2]);
        fprintf(stdout, "n: %f %f %f\n", n[0], n[1], n[2]);
        exit(EXIT_FAILURE);
    }
    );

    free(n);

    return height;


}


double
BuscarAltura(const double *punto, const struct T_Malla *malla)
{
    // size_t triangle = getTriangle(punto[0], punto[1], malla, &(malla->terreno.situation));
    size_t triangle = getTriangle_QT(punto[0], punto[1],
            (const double **)malla->nodo, (const size_t **)malla->terreno.cara,
            malla->terreno.Num_caras, &(malla->terreno.situation));

    if (triangle == malla->terreno.Num_caras+1){
        fprintf(stderr, "No se ha localizado el triangulo\n");
        exit(EXIT_FAILURE);
    }

    size_t idx_nodo1 = malla->terreno.cara[triangle][0];
    size_t idx_nodo2 = malla->terreno.cara[triangle][1];
    size_t idx_nodo3 = malla->terreno.cara[triangle][2];

    const double punto1[3] = {
        malla->nodo[idx_nodo1][0],
        malla->nodo[idx_nodo1][1],
        malla->nodo[idx_nodo1][2]};

    const double punto2[3] = {
        malla->nodo[idx_nodo2][0],
        malla->nodo[idx_nodo2][1],
        malla->nodo[idx_nodo2][2]};

    const double punto3[3] = {
        malla->nodo[idx_nodo3][0],
        malla->nodo[idx_nodo3][1],
        malla->nodo[idx_nodo3][2]};

    double altura = getHeight(punto1, punto2, punto3, punto);

    return altura;
}
