#ifndef TOPOGRAFIA_H
#define TOPOGRAFIA_H

#include "interpolacion.h"
#include "lectura.h"

#include <stdbool.h>
#include <stdio.h>

/**
 * @file topografia.h
 * @brief Funciones y estructuras para gestionar la topografia
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @brief Lee el fichero de topografia y rellena con sus datos
 *        una estructura de tipo S_Topografia
 * @param[in] fich Nombre del fichero donde esta la topografia
 * @param[out] topo puntero a una estructura de tipo S_Topografia
 *        donde se almacenaran los datos
 */
void 
leer_topo(const char *fich, struct T_Topografia *topo);

/**
 * @brief Función que interpola la rugosidad del mapa de topografia
 * @param[in] x,y Coordenada X Y del punto a interpolar
 * @param[in] topo Estructura que guarda los datos del mapa de topografia
 * @return Valor de la rugosidad interpolada en el punto (x,y)
 */
double InterpolaTopo(const double *x, const double *y, const struct T_Topografia *topo);

/**
 * @brief Función que lee los datos de una estacion desde un fichero
 * @param[inout] fp Fichero a leer
 * @param[out] est Estructura de estación donde guardar los datos
 * @param[in] tipo El viento está definido con componentes o módulo/ángulo
 * @param[in] fich Nombre del fichero a Leer
 * @param[inout] line_number Número de la línea leída del fichero
 * @return Número de muestreos del fichero
 */
void LeerEstacion(FILE *fp, struct T_Estacion *est, const size_t tipo, const char *fich, size_t *line_number);

/**
 * @brief Función que lee todos los datos de las distintas estaciones
 * @param[in] fich Puntero al nombre de ficheros de las estaciones
 * @param[in] altEstaciones Altura sobre el suelo de las estaciones
 * @param[in] malla Estructura de la malla
 * @param[out] Est Estructura de estaciones donde se guardan los datos leídos
 * @return Número de ficheros leídos
 */
size_t LeerLasEstaciones(const char *fich, const double *altEstaciones,
                         struct T_NEstacion *Est, const struct T_Malla *malla);

/**
 * @brief Función que lee el viento interpolado para el tipo de problema interpolation
 * @param[in] fich_viento Fichero con el viento interpolado de entrada
 * @param[inout] malla Estructura con la malla donde se guarda el viento interpolado
 */
void leer_interpolacion(const char *fich_viento, struct T_Malla *malla);

/**
 * @brief Función que interpola horizontalmente el campo de viento entre las estaciones
 * @param[in] x,y Coordenada X Y del punto en el que encontrar el viento
 * @param[in] z_terr Altura del terreno en la proyección del punto
 * @param[in] epsilon Valor para ponderar
 * @param[in] alturaEstaciones Altura del suelo de las estaciones
 * @param[in] est Estructura donde se guardan los datos de todas las estaciones
 * @param[out] V Vector del viento interpolado
 */
void InterpolacionHorizontal(const double *x, const double *y, const double *z_terr,
                             const double *epsilon,
                             const double *alturaEstaciones,
                             const struct T_NEstacion *est,
                             struct T_Vector3D *V);

/**
 * @brief Función que dada la altura de la estación con todas sus variables, encuentra la velocidad en la altura del punto
 * @param[in] alturaPunto Altura del punto a interpolar
 * @param[in] alturaRugosidad Longitud de rugosidad
 * @param[in] datmos Variable con los datos atmosféricos
 * @param[in] alturaEstacion Altura de la estación con datos conocidos
 * @param[out] V Campo de viento en la altura alturaPunto
 */
void InterpolacionVertical(const double alturaPunto, const double alturaRugosidad,
                           const struct T_DatosAtmos *datmos,
                           const double alturaEstacion, struct T_Vector3D *V);


/**
 * @brief Función que interpola el campo de viento en un punto a partir de las estaciones
 * @param[in] x,y,z Coordenada X Y Z del punto
 * @param[in] z_terr Altura del terreno en la proyección del punto
 * @param[in] datmos Estructura con los datos atmosféricos
 * @param[in] nest Estructura con todos los datos de las estaciones de viento
 * @param[in] topo Estructura con el mapa de la topografia
 * @return Devuelve un vector con las 3 componentes del viento interpolado
 */
struct T_Vector3D InterpolacionViento(const double x, const double y, const double z, const double z_terr,
                                        const struct T_DatosAtmos *datmos,
                                        const struct T_NEstacion *nest,
                                        const struct T_Topografia *topo);

/**
 * @brief Función que calcula el valor de fi para usar en la extrapolación vertical
 * @param[in] pasquill Tipo de estabilidad de pasquill
 * @param[in] z0 Valor de la rugosidad
 * @param[in] z Altura sobre el suelo del punto
 * @return El valor de fi
 */
double calculaFI(const int pasquill, const double *z0, const double *z);

/**
 * @brief Función que calcula la inversa de la longitud de Monin-Obukov
 * @param[in] estabilidad Tipo de estabilidad de Pasquill
 * @param[in] z0 valor de la rugosidad
 * @return El valor de la inversa de la longitud de Monin-Obukov
 */
double CalculaInversaLongitudMO(const int estabilidad, const double *z0);

/**
 * @brief Función que dado un punto encuentra la altura del punto de la superfície del terreno
 * @param[in] punto Punto que buscar
 * @param[in] malla Malla en la que encontrar la altura de la superficie
 * @return Altura del punto de la superficie
 */
double
BuscarAltura(const double *punto, const struct T_Malla *malla);

/**
 * @brief Función que interpola verticalmente la estación i-esima para que todos los datos esten a la misma altura altEstaciones
 * @param[in] Datmos Variable con los datos atmosféricos
 * @param[in] alturaAnemometro Altura actual de la estación
 * @param[out] Est Variable con los datos de las estaciones
 * @param[in] altEstaciones Altura a la que calcular el viento
 * @param[in] i Número de estación
 * @param[in] topo Variable con la longitud de rugosidad
 */
void correccionAlturaEstaciones(const struct T_DatosAtmos *Datmos,
                                const double alturaAnemometro,
                                struct T_NEstacion *Est,
                                const double* altEstaciones,
                                const size_t i, const struct T_Topografia *topo);

/**
 * @brief Función que escribe en pantalla el formato del fichero de topografia
 */
void uso_topo();

/**
 * @brief Función que dado 3 puntos te devuelve su vector normal
 * @param[in] p1,p2,p3 Coordenadas de los 3 puntos
 * @param[out] n Vector normal
 */
void getNormalVector(const double *p1, const double *p2, const double *p3, double *n);

/**
 * @brief Función que interpola la altura en punto, con los datos de punto1 punto2 y punto3
 * @param[in] punto1,punto2,punto3 Coordenadas de los puntos dato
 * @param[in] punto Coordenada del punto a interpolar la altura
 * @return Altura del punto
 */
double getHeight(const double *punto1, const double *punto2, const double *punto3, const double *punto);

#endif //TOPOGRAFIA_H
