#include "viento.h"

#include "calculo.h"
#include "comparisons.h"
#include "debugInfo.h"

#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>


/**
 * @file viento.c
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

void MeInterpolesElViento(struct T_Malla *malla, const struct T_DatosAtmos *Datmos,
                          const struct T_NEstacion *todas_est, const struct T_Topografia *topo)
{


    struct T_Vector3D Vh;

    for(size_t i = 0; i < malla->Num_tetra; i++) {
        // calcula y almacena el viento inicial en en centro del tetraedro
        Vh = InterpolacionViento(malla->cen_tetra[i][0], malla->cen_tetra[i][1],
                malla->cen_tetra[i][2],
                malla->z_terr[i],
                Datmos, todas_est,
                topo);
        malla->vini[i][0] = Vh.x;
        malla->vini[i][1] = Vh.y;
        malla->vini[i][2] = Vh.z;

    }
} //MeInterpolesElViento

void MeCalculesElViento(struct T_Malla *malla, const struct T_DatosAtmos *Datmos)
{
    double *viento;          // solucion del sistema
    double b[4];             // vector elemental
    struct T_Vector_Nodos *matriz;
    struct T_MElemental *m_eltal;   // matriz elemental


    // Reserva memoria para la matriz elemental de 4x4
    m_eltal = CrearMatrizElemental(4);

    // reserva memoria para el vector de listas encadenadas
    // necesario para ensamblar la matriz morse
    matriz = CrearVectorDeListas(malla->Num_nodos);

    for (size_t i = 0; i < malla->Num_tetra; i++) {
        // calcula la matriz y el vector elemental correspondientes
        // al tetraedro
        Elemento(i, malla, m_eltal, b, &(Datmos->alpha));

        // los ensambla en la lista encadenada
        EnsamblarMatrizElemental(matriz, m_eltal, malla, b);
    }

    // Generar el sistema de ecuaciones y resolverlo usando BiCGSTAB
    viento = ResolverSistema(matriz, malla->Num_nodos, Datmos->metodo,
                             Datmos->precond, Datmos->w_prec);
    LiberarVectorDeListas(matriz, malla->Num_nodos);
    // Calcula el viento final
    CalculaViento(malla, viento, &(Datmos->alpha));
    // Liberar memoria
    free(viento);
    DestruirMatrizElemental(m_eltal);

} // MeCalculesElVientoWRF

void MeInterpolesElVientoHARM(struct T_Malla *malla, const struct T_DatosAtmos *Datmos)
{
    time_t crono1;
    time_t crono2;
    static char *nfun = "MeInterpolesElVientoHARM";

    crono1 = time((time_t *) NULL);
    for (size_t i = 0; i < malla->Num_tetra; i++) {
        double vx, vy, vz;
        // calcula y almacena el viento inicial en en centro del tetraedro
        Interpola7VecinosHARM(Datmos->H,
                              &malla->cen_tetra[i][0], &malla->cen_tetra[i][1],
                &malla->cen_tetra[i][2],
                &vx, &vy, &vz);
        DEBUG_INFO(
                    fprintf(stdout, "%s: coord: %f %f %f   interp: %f %f %f\n", nfun,
                           malla->cen_tetra[i][0], malla->cen_tetra[i][1],
                malla->cen_tetra[i][2],
                vx, vy, vz);
        );

        malla->vini[i][0] = (double) vx;
        malla->vini[i][1] = (double) vy;
        malla->vini[i][2] = (double) vz;
    }
    crono2 = time((time_t *) NULL);
    fprintf(stdout, "%s: tiempo de interpolacion del viento con HARMONIE (seg) = %ld\n",
           nfun, crono2 - crono1); fflush(stdout);

}// MeInterpolesElVientoHARM


struct T_Vector3D VientoEnPunto(const struct T_Malla *malla, const struct T_Topografia *topo,
                                const struct T_DatosAtmos *Datmos,
                                const struct T_NEstacion *estaciones, const size_t tet,
                                const double x, const double y, const double z,
                                const double z_terr, const char *file_format,
                                struct T_Vector3D *vini)
{
    double l[3];
    struct T_Vector3D Vh;

    for (size_t k=0; k<3; k++){
        l[k] = malla->wind[tet][k] - malla->vini[tet][k];
    }

    // Calcula vini en el nodo en cuestion
    if (strcmp(file_format, "forecast") == 0) {
        Interpola7VecinosHARM(Datmos->H, &x, &y, &z, &(vini->x), &(vini->y), &(vini->z));
    } else {
        *vini = InterpolacionViento(x, y, z,
                                    z_terr,
                                    Datmos, estaciones, topo);
    }
    Vh.x = vini->x + l[0];
    Vh.y = vini->y + l[1];
    Vh.z = vini->z + l[2];
    return(Vh);
}// VientoEnPunto

double ConvertWindCart2Met(const double vx, const double vy)
{
    const double K=180.0/acos(-1.0);
    double a;

    if (greatThanOrEqualToZero(&vx) && greatThanZero(&vy)){
        a = atan(vx/vy)*K;
    } else if (lessThanOrEqualToZero(&vx) && greatThanZero(&vy)){
        a = 360+atan(vx/vy)*K;
    } else if (!equalToZero(&vy)){
        a = 180+atan(vx/vy)*K;
    } else if (greatThanZero(&vx)){
        a = 90;
    } else if (lessThanZero(&vx)){
        a = 270;
    } else{
        a = 0;
    }

    int angle = ((int)round(a)+180) % 360;
    return (double)angle;
}
