#ifndef VIENTO_H
#define VIENTO_H

#include "wind3D.h"
#include "topografia.h"
#include "lectura.h"

/**
 * @file viento.h
 * @brief Módulo con las funciones relativas al cálculo de viento
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @brief Función que interpola el viento en toda la malla
 * @param[in] malla Estructura de malla donde se guarda el valor del viento interpolado
 * @param[in] Datmos Estructura de datos atmosféricos
 * @param[in] todas_est Estructura con todos los datos de las estaciones de viento
 * @param[in] topo Estructura con el mapa de topografia
 */
void
MeInterpolesElViento(struct T_Malla *malla, const struct T_DatosAtmos *Datmos,
           const struct T_NEstacion *todas_est, const struct T_Topografia *topo);


/**
 * @brief Función que calcula el campo final de viento en toda la malla
 * @param[inout] malla Estructura de malla donde se guarda el campo de viento resultante
 * @param[in] Datmos Estructura de datos atmosféricos
 */
void
MeCalculesElViento(struct T_Malla *malla, const struct T_DatosAtmos *Datmos);

/**
 * @brief Función que calcula el viento resultante en un punto
 * @param[in] malla Estructura de malla
 * @param[in] topo Estructura de mapa de topografia
 * @param[in] Datmos Estructura con los datos atmosféricos
 * @param[in] estaciones Estructura con los datos de las estaciones de medida de viento
 * @param[in] tet Tetraedro al que pertenece el punto
 * @param[in] x,y,z Coordenada X Y Z del punto
 * @param[in] z_terr Valor de la altura del terreno en la proyección del punto
 * @param[in] file_format Formato de adquisición de datos del viento (pronóstico o estaciones)
 * @param[out] vini Velocidad del viento interpolada en el punto
 * @return Devuelve un vector con el viento resultante en el punto
 */

struct T_Vector3D
VientoEnPunto(const struct T_Malla *malla, const struct T_Topografia *topo,
          const struct T_DatosAtmos *Datmos,
          const struct T_NEstacion *estaciones, const size_t tet,
          const double x, const double y, const double z, const double z_terr, const char *file_format, struct T_Vector3D *vini);

/**
 * @brief Función que interpola el campo de viento en todos los nodos
 *        de la malla a partir de los datos del Harmonie
 * @param[out] malla Estructura de malla donde se guarda el campo de viento interpolado
 * @param[in] Datmos Estructura con los datos atmosféricos del harmonie
 */
void MeInterpolesElVientoHARM(struct T_Malla *malla, const struct T_DatosAtmos *Datmos);

/**
 * @brief Función que dado las componentes de la velocidad X Y, calcula el ángulo de viento (0 N, 90 E, 180 S, 270 W)
 * @param[in] vx,vy Componentes de la velocidad X Y
 * @return Ángulo del viento
 */
double ConvertWindCart2Met(const double vx, const double vy);

#endif //VIENTO_H
