#include "wind3D.h"

#include "avs.h"
#include "capas.h"
#include "caras_ext.h"
#include "localizador.h"
#include "puntosSalida.h"
#include "viento.h"

#include <ctype.h>
#include <getopt.h>
#include <stdlib.h>
#include <string.h>

/**
 * @file wind3D.c
 * @author Eduardo Rodríguez Barrera
 * @author Gustavo Montero García
 * @author José María Escobar Sánchez
 * @author Rafael Montenegro Armas
 * @author Albert Oliver Serra
 * @copyright Copyright (C) 2012 ULPGC
 * Universidad de Las Palmas de Gran Canaria
 * Institute of Intelligent Systems and 
 * Numerical Applications in Engineering - SIANI \n
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. \n
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. \n
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

void creaMalla(const char *f_malla, struct T_Malla *malla)
{
    LeerMalla(f_malla, malla);
    CorreccionNumerosReferencia(malla);

    getLocation_quadTree(malla->terreno.x_min, malla->terreno.y_min, malla->terreno.z_min,
                         malla->terreno.x_max, malla->terreno.y_max, malla->terreno.z_max,
                         (const double **)malla->nodo, (const size_t **)malla->terreno.cara,
                         malla->terreno.Num_caras, &(malla->terreno.situation));


    if (malla->puntosSalida || malla->tieneCapas || malla->puntosControl)
    {

        getLocation_tetrahedron(30,30,20,
                                malla->terreno.x_min, malla->terreno.y_min, malla->terreno.z_min,
                                malla->terreno.x_max, malla->terreno.y_max, malla->terreno.z_max,
                                malla->nodo, malla->tetra, malla->Num_tetra, &malla->situation);
    }


    for (size_t i = 0; i < malla->Num_tetra; i++) {
        double x, y, z;

        x = y = z = 0.0;
        for (size_t j=0; j<4; j++) {
            x += malla->nodo[malla->tetra[i][j]][0];
            y += malla->nodo[malla->tetra[i][j]][1];
            z += malla->nodo[malla->tetra[i][j]][2];
        }
        x /= 4.0;  y /= 4.0; z /= 4.0;
        malla->cen_tetra[i][0] = x;
        malla->cen_tetra[i][1] = y;
        malla->cen_tetra[i][2] = z;
        malla->z_terr[i] = BuscarAltura(malla->cen_tetra[i], malla);
    }

}


void leerViento(const char *file_format, struct T_DatosAtmos *Datmos, struct T_NEstacion *todas_est,
                struct T_Topografia *topo, struct T_Malla *malla)
{
    if (strcmp(file_format,"forecast") == 0) {

        LeerVientoHARMONIE(Datmos->fich_viento, Datmos->H);

    } else if (strcmp(file_format, "stations") == 0) {

        if (!Datmos->tieneRugosidad){
            leer_topo(Datmos->fich_topo, topo);
        }

        LeerLasEstaciones(Datmos->fich_viento, &Datmos->altest, todas_est, malla);
    } else if (strcmp(file_format, "interpolation") == 0){
        leer_interpolacion(Datmos->fich_viento, malla);
    }
}

void interpolacionViento(const struct T_DatosAtmos *Datmos, struct T_Malla *malla,
                         const struct T_Topografia *topo, struct T_NEstacion *todas_est,
                         const char *file_format)
{
    if (strcmp(file_format,"forecast") == 0) {
        // Interpola el viento a partir de datos del HARMONIE y calcula el
        // viento segun MMC
        MeInterpolesElVientoHARM(malla, Datmos);
    } else if (strcmp(file_format, "stations") == 0) {

        if (todas_est->noMismaAltura)
        {
            for (size_t i = 0; i < todas_est->NEST; ++i)
            {
                if (todas_est->N[i]->noMismaAltura)
                {
                    correccionAlturaEstaciones(Datmos, todas_est->N[i]->z_terr, todas_est, &Datmos->altest, i, topo);
                }
            }
        }

        MeInterpolesElViento(malla, Datmos, todas_est, topo);
    }

}

void 
calculoViento(struct T_Malla *malla, const struct T_DatosAtmos *Datmos)
{
    MeCalculesElViento(malla, Datmos);
}

void 
escribirResultados(const struct T_Malla *malla)
{
    GrabarVientoAVS("wind_field.inp", malla);
    GrabarModuloVientoAVS("wind_escalar.inp", malla);
}

void 
libera_memoria(char *fich_param, char *file_format, struct T_Malla *malla,
               struct T_DatosAtmos *Datmos, struct T_Topografia *topo,
               struct T_NEstacion *todas_est)
{
    for (size_t i = 0; i < malla->Num_nodos; i++){
        free(malla->nodo[i]);
    }
    free(malla->nodo);

    for (size_t i = 0; i < malla->Num_tetra; i++){
        free(malla->tetra[i]);
        free(malla->cen_tetra[i]);
        free(malla->vini[i]);
        free(malla->wind[i]);
    }

    for (size_t i = 0; i < malla->terreno.Num_caras; i++){
        free(malla->terreno.cara[i]);
    }
    free(malla->terreno.cara);

    freeQT(&(malla->terreno.situation));

    free(malla->tetra);
    free(malla->cen_tetra);
    free(malla->z_terr);
    free(malla->vini);
    free(malla->wind);

    free(malla->condicion);

    if (malla->puntosSalida || malla->tieneCapas || malla->puntosControl){
        for (size_t i=0; i<malla->situation.nx; i++){
            for (size_t j=0; j<malla->situation.ny; j++){
                for (size_t k=0; k<malla->situation.nz; k++){
                    free(malla->situation.location[i][j][k]);
                }
                free(malla->situation.location[i][j]);
                free(malla->situation.size[i][j]);
            }
            free(malla->situation.location[i]);
            free(malla->situation.size[i]);
        }
        free(malla->situation.location);
        free(malla->situation.size);
    }

    if (malla->puntosSalida){
        free(malla->fichPuntosSalida);
    }

    if (malla->puntosControl){
        free(malla->fichPuntosControl);
    }

    free(Datmos->fich_viento);
    if (!Datmos->tieneRugosidad){
        free(Datmos->fich_topo);
    }
    free(Datmos->fich_malla);

    if (strcmp(file_format, "forecast") == 0){
        free(Datmos->H->latitud);
        free(Datmos->H->longitud);
        free(Datmos->H->altura);
        free(Datmos->H->data_u);
        free(Datmos->H->data_v);
        free(Datmos->H->data_w);
    }

    if (strcmp(file_format, "stations") == 0){
        if (!Datmos->tieneRugosidad){
            for (size_t i=0; i < topo->nx; i++){
                free(topo->z[i]);
                free(topo->z0[i]);
            }
            free(topo->z);
            free(topo->z0);
        }

        for (size_t i=0; i < todas_est->NEST; i++){
            free(todas_est->N[i]);
        }
        free(todas_est->N);
    }

    free(fich_param);
    free(file_format);

}

